!This module contains the routines to compute the likelihood from observerables
!update this module with new observational data
!SuperBayes Package - by Roberto Ruiz de Austri (rruiz@delta.ft.uam.es) and Roberto Trotta (rxt@astro.ox.ac.uk)
!This version October 2007 

! Revision 1.13 2008/01/12 
! Bug-fix in the smoothed 2D like for the Higgs mass routine qtrap(..)
! which for a very small mass givelnlike = 0 and Leftlnlike = 0
! The ifc v8 takes Leftlnlike/lnlike to be 0 then working fine  
! while v9 and v10 give NaN which makes the loop go on
! forever. To avoid it, at line 500, the extra condition (Leftlnlike == 0.d0) 
! added. Same for Rightlnlike at line 512.  

module CalcLike
  use Random
  use settings
  use ParamDef
  use likedata 

  implicit none
  
  logical :: Init_data

  !Global variables needed by many functions/subroutines
  real(8) :: mh_high, G_mh, G_zeta, G_sigma_zeta, G_tau_mh, G_tau_zeta
  integer:: numav

contains
  
function GetLogLike(Params) 
  !Returns -Ln(Likelihood) = chi^2/2
  Type(ParamSet)::  Params 
  Type(Input_Params):: MSSM
  Type(Nuisance_Params):: NuisP
  Type(Susy_spec):: SUSYSpec
  Type(DS_Output) :: DSO
  Type(SoftOut) :: OutSoft 
  Type(Bdecay_Out) :: Bdecay
  real GetLogLike
  real dum(1,1)
  logical :: ErrorsFromSoftSusy
  Type(Prodxbr_Out) :: Prodxbr

  logical, parameter :: DebugSampler = .false.

  numav = num
  if (restart) numav = num - num_at_restart
  !this compensates for the large num at restart

  if ((any(Params%P > Scales%PMax) .or. any(Params%P < Scales%PMin)) .and. .not. postproc) then
     !proposal outside prior range
     GetLogLike = logZero
     if (Feedback > 2) write(*,*) 'out of bounds'
     return
  end if

  if (.not. DebugSampler) then
        !either MCMC or if postprocessing request to do theory again
        !converts Parms to MSSM
        call ParamsToMSSMParams(Params%P, MSSM, NuisP)

        !calls SoftSusy to get the spectrum
        if (timing) call StartTiming(Hertz,Begin_Clock)       
        call GetSoftSusySpectrum(MSSM, NuisP, SUSYSpec, OutSoft, DSO%Err)
        
        if (timing) then
           call StopTiming(Begin_Clock, End_Clock, Hertz, Elapsed_Time, ms)
           call ComputeAverage(softsusy_av, softsusy_ssq,ms,numav)
        end if
       

        ErrorsFromSoftSusy = ErrorsPresent(DSO%Err)
        
        if (Feedback > 2) Call DebugSoftErrorOutput(DSO%Err)   
        !calls DS and Feynhiggs to predict the quantities of interest
        if (.not. ErrorsFromSoftSusy) then

           !sets up ID inputs
           if(GFlags%ID_predict) then
              if (.not. allocated(DSO%ID_in%hmodel)) allocate (DSO%ID_in%hmodel(num_hm))
              DSO%ID_in = GIDin          
           end if
           !Prodxbr will contain on output the predictions by Feynhiggs

           call GetPredictions(MSSM, NuisP, OutSoft, DSO, Prodxbr, BDecay)

           !warning has been issued by DS
           !if Feedback > 3, this has already been written
           if (DSO%PrintOutParams .and. Feedback < 3) call PrintOutModelParams(MSSM, NuisP)
           !Special Treatment for problems coming from oh2:
           if (DSO%err%oh2 .and. Feedback < 3) then
              write(*,*) 'Warning: problem with oh2 from this point:'
              call PrintOutModelParams(MSSM, NuisP)         
           end if
           
           !global variable, saves the results from the current Trial point
           !if the Current is accepted, than the Previous is written
           !in WriteParams
           !the variables are defined in paramdef.f90 
           CurrentOutputP = DSO 
           CurrentSusySp  = SUSYSpec 
           CurrentSoftOut = OutSoft
           CurrentProdxbr = Prodxbr
           CurrentBdecay  = Bdecay
           !Prodxbr at the moment not passed to calclike since not used for data comparison
           
           if (Feedback > 2) write(*,*) '... Computing likelihood ... '
           if (timing) call StartTiming(Hertz,Begin_Clock)       
           GetLogLike = GetLogLikePost(SUSYSpec, NuisP, DSO, MSSM, OutSoft, BDecay)
           if (timing) then
              call StopTiming(Begin_Clock, End_Clock, Hertz, Elapsed_Time, ms)
              call ComputeAverage(like_av, like_ssq,ms,numav)
           end if
           
        else
           !if errors in SoftSusy, point is non-physical
           GetLogLike = LogZero
           return
        end if
  else !debug purposes
     GetLogLike = FakeGetLogLike(Params)
  endif !DebugSampler
  
  if (GetLogLike /= LogZero) GetLogLike = GetLogLike/Temperature

end function GetLogLike


function FakeGetLogLike(Params) 
   !Only for debug purposes
  Type(ParamSet)::  Params  
  real FakeGetLogLike
  real :: s1, s2, rho, mu1, mu2
  real :: Pd
  if (any(Params%P > Scales%PMax) .or. any(Params%P < Scales%PMin)) then
     FakeGetLogLike = logZero
     write(*,*) 'Out of Bounds in Fake Params'
     return
  end if
  if (.false.) then
     mu1 = 4000d0
     mu2 = 1000d0
     s1 = 1500d0
     s2 = 25d0
     rho = 0.5
     FakeGetLogLike = 1/(2*(1 - rho**2))* &
          & ( (mu1-Params%P(1))**2/(s1**2) + (mu2-Params%P(2))**2/(s2**2)&
          & - 2*rho*(mu1-Params%P(1))*(mu2-Params%P(2))/(s1*s2))
  end if
  !simple 1D case
  !P1 is the input, Pd a fake derived param
  !I have data on the derived param only
  Pd = (LOG10(Params%P(1)))
  mu1 = 3.3d0
  s1 = 0.2d0
  mu2 = 5.0d0
  s2 = 2.0d0
  rho = 0.5
  !correlated params
  FakeGetLogLike = 1/(2*(1 - rho**2))* &
       & ( (mu1-Pd)**2/(s1**2) + (mu2-Params%P(3))**2/(s2**2)&
       & - 2*rho*(mu1-Pd)*(mu2-Params%P(3))/(s1*s2))
  
  !the derived parm is saved in P(2)
  Params%P(2) = Pd
  
end function FakeGetLogLike
    
  function GetLogLikePost(SusySp, NuisP, OutP, InP, OuS, Bd)
    !here the -Ln(like) = chi^2/2 is computed
    !edit this function to include new data etc
    real GetLogLikePost, NewlogLike
    Type(Susy_spec) :: SusySp
    Type(Nuisance_Params):: NuisP
    Type(DS_Output) :: OutP
    Type(SoftOut) :: OuS
    Type(Input_Params) :: InP
    Type(BDecay_Out) :: Bd

    integer error

    !checks for softsusy errors already done
    !here only checks the error from DS
    if (OutP%err%unphys .eq. 1) then
       GetLogLikePost = logZero
       if (Feedback > 2) write(*,*) 'DS error'
       return
    else 
       !LogLike from Data
       if (Use_CDM) GetLogLikePost = GetLogLikePost + GetLogLikeCDM(OutP)
       if (Feedback > 2) write(*,'(A,E15.7)')    '-lnlike CDM         = ', GetLogLikePost
       if (Use_Nuisance) then
          NewLogLike = GetLogLikeNuisance(NuisP)
          GetLogLikePost = GetLogLikePost + NewLogLike
          if (Feedback > 2) write(*,'(A,E15.7)') '-lnlike nuisance    = ', NewLogLike
       end if
       !warning, DD GetLogLike must come after CDM LogLike if LSP not the only CDM component
       !(rescaling of cross sections is done in CDM LogLike)
       if (Use_DD) call DoStop('Direct Detection bounds not implemented!')

       if (Use_ID) call DoStop('Indirect Detection bounds not implemented!')

       if (Use_LEP) then
          if (use_data .eq. current_data) then
             NewLogLike = GetLogLikeMHiggs(SusySp, InP, OuS)
          else if (use_data .eq. synthetic_data) then
             NewLogLike = GetLogLikeFromData(MHiggs, SusySp%mlh)
          end if
          GetLogLikePost = GetLogLikePost + NewLogLike
          if (Feedback > 2) write(*,'(A,E15.7)') '-lnlike M_higgs     = ', NewLogLike
          NewLogLike = GetLogLikeMSparticles(SusySp) 
          if (Feedback > 2) write(*,'(A,E15.7)') '-lnlike sparticles  = ', NewLogLike
          GetLogLikePost = GetLogLikePost + NewLogLike         
       end if


       if (Use_Anomalous_Mu) then
          NewLogLike = GetLogLikeFromData(AnomalousMu,  OutP%coll%gm2amp*1E10) 
          GetLogLikePost = GetLogLikePost + NewLogLike 
          if (Feedback > 2) write(*,'(A,E15.7)') '-lnlike g-2         = ', NewLogLike
       end if

       if (Use_bsgamma) then
          !checks the relative sign of the amplitude b -> s gamma between MSSM and SM
          !if (OutP%coll%sgn_MSSMoSM  is zero they have opposite sign and the point must be rejected, RT Oct 05
          if (OutP%coll%sgn_MSSMoSM .eq. 0) then
             GetLogLikePost = LogZero
          else
             NewLogLike = GetLogLikeFromData(Bsgamma, OutP%coll%bsgamma*1E4)
             GetLogLikePost = GetLogLikePost +  NewLogLike
             if (Feedback > 2) write(*,'(A,E15.7)') '-lnlike Bs -> gamma = ', NewLogLike
          end if
       end if

       if (Use_Bsmumu) then
          NewLogLike = GetLogLikeFromData(Bsmumu, OutP%coll%brbsmm*1E7)
          GetLogLikePost = GetLogLikePost +  NewLogLike
          if (Feedback > 2) write(*,'(A,E15.7)') '-lnlike Bs -> mumu  = ', NewLogLike
       end if

       !Now EWPOs from ds instead
       if (Use_Mass_W) then
          NewLogLike = GetLogLikeFromData(MW, OutP%coll%mw) 
          GetLogLikePost = GetLogLikePost +  NewLogLike
          if (Feedback > 2) write(*,'(A,E15.7)') '-lnlike M_W         = ', NewLogLike
       end if

       if (Use_Weak_Mixing) then
          NewLogLike = GetLogLikeFromData(SinEff, OutP%coll%sweff2)
          GetLogLikePost = GetLogLikePost +  NewLogLike
          if (Feedback > 2) write(*,'(A,E15.7)') '-lnlike sineff      = ', NewLogLike
       end if

       if (Use_Bdecay) then
          NewlogLike = GetLogLikeFromData(BdecayData, Bd%delMB)
          GetLogLikePost = GetLogLikePost +  NewLogLike
          if (Feedback > 2) write(*,'(A,E15.7)') '-lnlike Bdecay      = ', NewLogLike
       end if
       
    end if
  
end function GetLogLikePost

function ErrorsPresent(err)
  logical ErrorsPresent
  Type(Error_Out) :: err

  if ((err%muSqWrongSign.eq.1).or.(err%b.eq.1).or. &
       & (err%noRhoConvergence.eq.1).or.(err%tachyon.eq.1).or.&
       & (err%nonperturbative.eq.1).or.(err%noMuConvergence.eq.1).or.&
       & (err%higgsUfb.eq.1).or.(err%irqfp.eq.1).or.&
       & (err%badConvergence.eq.1)) then
     ErrorsPresent = .true.
  else
     ErrorsPresent = .false.
  end if

end function ErrorsPresent

function SmearedBound(y, x0, sigma, tau, lower)result(lnprob)
    !Returns the -ln(like) for a bound smeared with
    !theoretical and (if available) experimental error
    !According to Eq. (3.5) in 
    !R. Ruiz de Austri, R. Trotta and L. Roszkowski (2006)
    !JHEP 05 (2006) 002, hep-ph/0602028
    !------------- input --------------
    !   y      : value at which ln(like) is to be evaluated
    !   x0     : the lower/upper limit (ideally, the peak of the lower half-Gaussian,
    !            or else the 95% limit with very small sigma)
    !   sigma  : the 1sigma widht of the half-gaussian (experimental, when available)
    !   tau    : estimated 1sigma uncertainity of theoretical prediction
    !   lower  : = .true.  means x0 is a  lower bound
    !            = .false. means x0 is an upper bound
    !------------- output  --------------
    !   lnprob : -ln(like)
    !            normalized in such a way that -ln(like) = 0 well above/below the bound 
    !            (ie, where no constraints are present)
    real(8) :: y
    real(8) :: lnprob, x0, sigma, tau, tstar, exparg, Zfunc2
    logical :: lower
    real :: sgn

    if ((sigma < 0.0) .or. (tau < 0.0)) then
       write(*,*) 'sigma, tau = ', sigma, tau
       call DoStop('You are trying to use negative errors in SmearedBound!')
    end if
    if ((sigma .eq. 0.0) .and. (tau .eq. 0.0)) call DoStop('Both errors are 0 in SmearedBound!')
    !Eq. (3.5) is only valid for sigma>0, so need to put in a small number if sigma unavailable
    if ((sigma .eq. 0.0) .and. (tau>0.0)) sigma = tau/100d0
    if ((tau .eq. 0.0)) call DoStop('tau=0 in SmearedBound!')

    if (lower .eq. .true.) then
       sgn = 1.0d0
       else
       sgn = -1.0d0
    end if
    !See Eq. (3.6) in hep-ph/0602028
    tstar = (sigma/tau)*(sgn*(x0-y))/sqrt(sigma**2+tau**2)
    exparg = -0.5*(sgn*(y-x0))**2/(sigma**2+tau**2)
    Zfunc2 =  Z_func(sgn*(x0-y)/tau)
    !debug purposes
    !write(*,*) 'tstar ', tstar
    !write(*,*) 'Zfunc 1', Z_func(tstar)
    !write(*,*) 'Zfunc 2', Zfunc2
    !write(*,*) 'Exp arg', exparg

    if ( EXP(exparg) > 0d0) then 
       !retain the exp term
       !multiplied by sqrt(2Pi)sigma to ensure that it is normalized in such a way
       !that lnprob = 0 for no constraint
       lnprob = -LOG(sigma/sqrt((sigma**2+tau**2))*EXP(exparg)*(1-Z_func(tstar)) + Zfunc2)
    else
       !the first term in the LOG above is = 0. since Z_func(tstar) -> 1
       !this means we are either much above or much below the experimental limit
       !only the second term (Z_func2) is important
       if (Zfunc2 .eq. 0d0) then
          !we are in the excluded region
          lnprob = LogZero
       else if (Zfunc2 .eq. 1.0d0) then
          !no constraints in this region here
          lnprob = 0d0 
       else
          write(*,*) 'Something wrong in SmearedBound, Zfunc2 = ', Zfunc2
       end if
    
    end if

  end function SmearedBound


  function Z_func(lambda)
    !auxiliary function. Just the Gaussian erorrfc, rescaled
    !Z_func = 1/sqrt(2pi) int_lambda^infty exp(-1/2 t**2) dt
    real (8) :: Z_func, lambda
    Z_func = 0.5*errfc(lambda/sqrt(2.0))
  end function Z_func

  function errfc(x)
    !from Press et al p 164 
    !returns the complementary error fct with fractional precision of E-7 
    real(8) :: errfc, x,z,t 
    z = abs(x)
    t = 1.0/(1.0 + z/2.0)
    errfc = t*EXP(-z*z - 1.26551223 + t*(1.00002368 + t*(0.37409196+ &
        t*(0.09678418 + t*(-0.18628806 +t*(0.27886807 + t*(-1.13520398 + &
        t*(1.48851587 + t*(-0.82215223 +t*0.17087277)))))))))
    if (x .lt. 0.0) errfc = 2.-errfc
    return
  end function errfc

!-------------------------------------------
!    Generic multi-purpose likelihood
!------------------------------------------
  function GetLogLikeFromData(Datum, theory)result(lnlike)
    real :: lnlike
    real(8) :: theory
    real(8) :: tau
    Type(LikeDatum) :: datum

    if (datum%tau_percent .eq. .true.) then
       tau = datum%tau*theory
    else
       tau = datum%tau
    end if

    !Gaussian likelihood here
    if (datum%datum_type .eq. Gaussian) then
       lnlike = 0.5*(theory-datum%mu)**2/(datum%sigma**2+tau**2)
    else if  (datum%datum_type .eq. UpperLimit) then
       lnlike = SmearedBound(theory, datum%mu, datum%sigma, tau,lower= .false.)
    else if  (datum%datum_type .eq. LowerLimit) then
       lnlike = SmearedBound(theory, datum%mu, datum%sigma, tau,lower= .true.)
    end if
  end function GetLogLikeFromData


!-------------------------------------------
!    CDM likelihood
!------------------------------------------
  function GetLogLikeCDM(OutP)result(lnlike)
    real  lnlike
    real  omch2, sigma, tau
    Type(DS_Output) :: OutP

    if (CDM_purely_LSP) then 
       lnlike= GetLogLikeFromData(CDM, OutP%dm%oh2)
    else
       !WMAP measurement is only an upper bound
       stop 'You need to check/change the bound in GetLogLikeCDM'
       !the bit below is untested, May 2007.
       if (OutP%dm%oh2 < CDM%mu) then
          !need to rescale the cross-sections
          OutP%dd%sigsi_nuc = OutP%dm%oh2/CDM%mu*OutP%dd%sigsi_nuc
          !if it's below the measurement, no constraints
          lnlike = 0d0
       else
          !if it's above, half a Gaussian
           lnlike=  GetLogLikeFromData(CDM, OutP%dm%oh2) 
       end if
    endif
  end function GetLogLikeCDM

!-------------------------------------------
!    Light Higgs likelihood (bound)
!------------------------------------------

  function GetLogLikeMHiggs(SUSYSpec, InP, OuS)result(lnlike)
    !RT rewritten June 06
    !This computes the likelihood in the (m_h, xi^2 plane)
    !uses LEPII bounds, does not assume SM-like couplings
    real(8)  lnlike, Rightlnlike, Leftlnlike
    real(8)  mh_low, IntMin, IntMax, LeftIntMin, LeftIntMax,  RightIntMin, RightIntMax
    Type(Susy_spec) :: SUSYSpec
    Type(Input_Params) :: InP
    Type(SoftOut) :: OuS

    integer, parameter :: tabmax = 90
    real(8) :: lnliketable(0:tabmax, 0:tabmax), G_zetatable(0:tabmax)
    real, parameter :: tolerance = 1E-3

    integer i,l, ii
    logical :: ExtendRight, ExtendLeft
    !for TESTING PURPOSES ONLY set t1 to .true.
    logical, parameter :: tl= .false.
    real(8) :: alpha, beta, SinBmASq, SinBmASq_exp, sbma2_intp,  SinBmASq_exp2, zeta_exp, zetath
    !RT revised Oct 04
    !this is a freely Bayesian reinterpretation of the frequentist
    !CL from hep-ex/0306033
    !here I compute the ratio of the couplings
    beta = atan(InP%tanb)
    alpha = OuS%higg%halpha
    SinBmASq = sin(beta-alpha)**2.d0 
    !This is the absolute lower bound (not used anymore)
    mh_low  = 91.0
    !This is the relaxed bound for SM-like couplings only
    mh_high = 114.4

    G_sigma_zeta = 0.0 !expe width not available, take it to be 0
    G_tau_zeta = 0.0 
    !theoretical error for zeta direction found to be irrelevant for couplings close
    !to SM-like (as in the CMSSM)
    !put it to zero. This condition can be relaxed, eg comment out
    !   G_tau_zeta.1*SinBmASq   
    !for a theoretical width of zeta prediction at 10% level
    G_tau_mh = 3.0 !3 GeV theoretical error on mh 
 
    !integrating out uncertainty due to m_higgs position
    !integral over zeta done analytically, over m_higgs numerically
    !Global quantities accessible from all subroutines/functions
    G_mh = SUSYSpec%mlh
    G_zeta = SinBmAsq


    !test purposes only
    if (tl) then
       write(*,*) ' D D D D D D D D D D D D'
       call CreateFile('liketest.dat', 99, 'formatted')
       call CreateFile('lnliketable_both10.dat', 97, 'formatted')
       call CreateFile('lnlikeaxis.dat', 95, 'formatted')
       lnliketable = 0d0
    end if

    do i=0,tabmax !loop over mh values
       if (tl) G_mh  = 90+i*0.5
        !do l=0,tabmax !loop over zeta values
           !if (tl) G_zeta = 0.05+l*0.02
       do l=0,1  
       if (tl) G_zeta = 0.99+l*0.01
           !write(*,*) 'computing mh, zeta',  G_mh ,  G_zeta
           IntMin = G_mh - 1.0*G_tau_mh
           IntMax = G_mh + 1.0*G_tau_mh
           if (G_tau_mh > 0d0) then
              !takes into account theoretical uncertainty in m_h
              call qtrap(HiggsLikeIntegrand, IntMin, IntMax, lnlike)
              if (tl) write(*,*) 'lnlike current', lnlike
              !try to extend the integral to see if converged
              ii=1
              ExtendLeft = .true.
              ExtendRight = .true.
              do
                 if (.not.(ExtendLeft) .and. .not.(ExtendRight)) exit
                 if (ii>10) stop 'Too many interations in MHiggs'
                 if (ExtendLeft) then
                    LeftIntMax = IntMin - (ii-1)*G_tau_mh
                    LeftIntMin = IntMin - ii*G_tau_mh
                    if (tl) write(*,*) 'Extending to L range ', LeftIntMin, LeftIntMax
                    call qtrap(HiggsLikeIntegrand, LeftIntMin, LeftIntMax, Leftlnlike)
                    if ((Leftlnlike/lnlike < tolerance) .or. & 
                        (Leftlnlike == 0.d0)) then
                       ExtendLeft = .false.
                    else
                       lnlike = lnlike + Leftlnlike
                    end if
                 end if
                 if (ExtendRight) then
                    RightIntMin = IntMax + (ii-1)*G_tau_mh
                    RightIntMax = IntMax + ii*G_tau_mh
                    if (tl) write(*,*) 'Extending to R range ', RightIntMin, RightIntMax
                    call qtrap(HiggsLikeIntegrand, RightIntMin, RightIntMax, Rightlnlike)
                    if (Rightlnlike/lnlike < tolerance) .or. &
                        (Rightlnlike == 0.d0)) then
                       ExtendRight = .false.
                    else
                       lnlike = lnlike + Rightlnlike
                    end if
                 end if
                 ii = ii+1
              end do
           else
           !no integration on m_higgs
              if (G_mh > mh_high) then
                 lnlike = 1d0 !no constraints, this is actually the like!
              else 
                 zeta_exp = sbma2_intp(G_mh,6)
                 zetath = G_zeta
                 if (G_tau_zeta> 0.001) then
                    !Smeared in zeta
                    !SmearedBound gives out -ln(prob) - need to exponentiate it
                    lnlike = EXP(-1.0d0*SmearedBound(zetath,zeta_exp, G_sigma_zeta, G_tau_zeta, lower=.false.))
                 else !No smearing at all
                    if (zetath > zeta_exp) then 
                       lnlike = 0d0 !ruled out - this is actually like,NOT lnlike yet
                    else
                       lnlike = 1d0 !allowed 
                    end if
                 end if !G_tau_zeta smearing
              end if !G_mh > mh_high
           end if !smearing on mhigss or not
           
           if (tl) then
              write(*,*) 'Prob', lnlike
              lnliketable(i,l) = lnlike
              G_zetatable(l) = G_zeta
              write(99,'(3E15.7)') G_mh, G_zeta, lnlike
           else
              !HiggsLikeIntegrand reurns the like, we need -ln(like)
              if (lnlike>0d0) then
                 lnlike = -LOG(lnlike)
              else
                 lnlike = LogZero
              end if
              !write(*,*) 'mh -lnlike and exiting:', lnlike
              exit !no loop
           end if
        end do !l to tabmax
        if (tl) then
           write(95,'(2E15.7)') G_mh, G_zetatable(i)
        else
           exit !no loop
        end if
     end do !ito tabmax loop
     if (tl) then
        do i=0,tabmax
           write(97,'(200E15.7)') lnliketable(i,:)/maxval(lnliketable(:,:))
        end do
        close(99)
        close(97)
        close(95)
        stop
     end if

! TESTING PURPOSES ONLY
!    write(*,*) ' D D D D D D D D D D D D'
!    call CreateFile('testobs.dat', 99, 'formatted')
!    do i=0,100
!       HiggsMass = 80+i*0.5
!       SinBmASq_exp = sbma2_intp(1.0d0*HiggsMass,6)
!       write(99,*) HiggsMass, SinBmASq_exp
!    end do
!    close(99)
!    Call DoStop('Done')


   end function GetLogLikeMHiggs


   function HiggsLikeIntegrand(mdum)result(ProbIntegrand)
     real(8) :: mdum, mth, zetath, ProbIntegrand
     real(8) :: zeta, zeta_exp, Like1, sbma2_intp
     !input: mth     = the theoretical predicted m_higgs 
     !       zetath  = is the value of zeta that goes with it (theoretically predicted)
     !       mdum    = is the "true" m_higgs value we want to integrate over
     !
     !output: value of the function to integrate over, ie
     !        L(d|zetath, m_dum) 1/(sqrt(2pi)sigma_m)exp(-0.5(mth-m_dum)^2/sigma_m^2) 
     !        This needs to be integrated over m_dum
     !sets this to global variables. so dont need to pass them
     mth = G_mh
     zetath = G_zeta

     !this is the like in zeta direction
     if (mdum > mh_high) then
        Like1 = 1.0
     else
        !write(*,*) 'calling interp'
        zeta_exp = sbma2_intp(mdum,6)
        if (G_tau_zeta> 0.001) then
           !SmearedBound gives out -ln(prob) - need to exponentiate it
           Like1 = EXP(-1.0d0*SmearedBound(zetath,zeta_exp, G_sigma_zeta, G_tau_zeta, lower=.false.))
        else !No smearing at all
           if (zetath > zeta_exp) then 
              Like1 = 0d0 !ruled out, this is the probability, not lnlike
           else
              Like1 = 1d0 !allowed, this is the probability, not lnlike 
           end if
        end if
     end if
     !write(*,*) 'zetath, zetaexp', zetath, zeta_exp
     ProbIntegrand = Like1/(sqrt(2*pi)*G_tau_mh)*EXP(-0.5*(mth-mdum)**2/G_tau_mh**2)
     !if (mdum < 78) write(*,'(A,3E15.7)') '>>>>(mh, zeta_exp(mh)) and prob ', mdum, zeta_exp, ProbIntegrand      
     
     return
   end function HiggsLikeIntegrand

!-------------------------------------------
!    Sparticles likelihood (bounds)
!------------------------------------------

   function GetLogLikeMSparticles(SUSYSpec)result(lnlike)
     real(8)  :: lnlike, percent_error
     real(8)  :: CharginoMass, MLSP
     real(8)  :: mstar, sigma, tau, mlsq 
     Type(Susy_spec) :: SUSYSpec
     integer :: i

     !taking 5% th error on all masses, sigma is unknown
     percent_error = 0.05
     sigma = 0.0

     CharginoMass = SUSYSpec%mch1
     MLSP = SUSYSpec%mneut1
     !-------------
     !chargino
     !-------------
     !write(*,*) 'Ch'
     if (SUSYSpec%msntau > 300 .and. (CharginoMass - MLSP) > 3) then
        mstar  = 103.5
     else
        mstar = 92.4
     end if
     tau = percent_error*CharginoMass
     lnlike = SmearedBound(CharginoMass,mstar,sigma, tau, lower = .true.)
     
    ! TESTING PURPOSES ONLY
    !write(*,*) ' D D D D D D D D D D D D'
    !call CreateFile('test_ch_mass_3.dat', 99, 'formatted')
    !do i=1,2000
    !   CharginoMass = i*0.1d0
    !   write(*,*) 'Now doing mass', CharginoMass
    !   mstar = 50d0
    !   sigma = 5d0
    !   tau = 0.05*CharginoMass      
    !   lnlike = SmearedBound(CharginoMass,mstar,sigma, tau, lower = .false.)
    !   write(99,*) CharginoMass, lnlike, exp(-1.0*lnlike)
    !end do
    !close(99)
    !Call DoStop('Done')



     !-------------
     !LSP
     !-------------
     !write(*,*) 'LSP'
     tau=percent_error*MLSP
     mstar = 50
     lnlike = lnlike + SmearedBound(MLSP,mstar,sigma,tau,lower = .true.)

     !-------------
     !sleptons
     !-------------
     !s electron R

     !write(*,*) 'Sl'
     tau=percent_error*SUSYSpec%mseR
     if (SUSYSpec%mseR - MLSP > 10) then
        mstar = 100
     else
        mstar = 73
     end if
     lnlike = lnlike + SmearedBound(SusySpec%mseR,mstar,sigma,tau, lower = .true.)
     
     !s muon R
     tau=percent_error*SUSYSpec%msmuR
     if (SUSYSpec%msmuR - MLSP > 10) then
        mstar = 95
     else
        mstar = 73
     end if
     lnlike = lnlike + SmearedBound(SusySpec%msmuR,mstar,sigma,tau,lower = .true.)
     
     !s tau 1
     tau=percent_error*SUSYSpec%mstau1
     if (SUSYSpec%mstau1 - MLSP > 10) then
        mstar = 87
     else
        mstar = 73
     end if
     lnlike = lnlike + SmearedBound(SusySpec%mstau1,mstar,sigma,tau,lower = .true.)
     
     !sneutrino (s nu tau is the lightest)
     tau=percent_error*SUSYSpec%msntau
     if (SUSYSpec%msntau - MLSP > 10) then
        mstar = 94
     else
        mstar = 43
     end if
     lnlike = lnlike + SmearedBound(SusySpec%msntau,mstar,sigma,tau,lower = .true.)    

     !-------------
     !squarks
     !-------------
     !write(*,*) 'Sq'
     !stop
     tau=percent_error*SUSYSpec%mst1
     if (SUSYSpec%mst1 - MLSP > 10) then
        mstar = 95
     else
        mstar = 65
     end if
     lnlike = lnlike + SmearedBound(SusySpec%mst1,mstar,sigma,tau,lower = .true.)
     
     !sbottom
     tau=percent_error*SUSYSpec%msb1
     if (SUSYSpec%msb1 - MLSP > 10) then
        mstar = 95
     else
        mstar = 59
     end if
     lnlike = lnlike + SmearedBound(SusySpec%msb1,mstar,sigma,tau, lower= .true.)
     
     !first 2 generations (d,u) and (s,c)
     !m_sqR < m_sqL
     mlsq = min(SUSYSpec%mscR, SUSYSpec%mssR, SUSYSpec%msuR, SUSYSpec%msdR)
     tau=percent_error*mlsq
     mstar = 318
     lnlike = lnlike + SmearedBound(mlsq,mstar,sigma,tau, lower= .true.)

     !-------------
     !gluino
     !-------------
     !write(*,*) 'glu'
     !TEMPORARY FIX ONLY!
     tau=abs(percent_error*SUSYSpec%mgluino)
     mstar = 233
     lnlike = lnlike + SmearedBound(SusySpec%mgluino,mstar,sigma,tau, lower= .true.)
     !RT March 06 updated

     if (lnlike > LogZero) lnlike = LogZero
     !to be consistent with everything else

   end function GetLogLikeMSparticles


!-------------------------------------------
!   Nuisance (SM) parameters
!------------------------------------------

  function GetLogLikeNuisance(NuisP)result(lnlike)
    real  lnlike
    Type(Nuisance_params) :: NuisP

    lnlike =  GetLogLikeFromData(Mtop, NuisP%mtop)
    lnlike = lnlike + GetLogLikeFromData(Mbottom, NuisP%mbottom)
    lnlike = lnlike + GetLogLikeFromData(alpha_S, NuisP%alphasmz)
    !the constraints are applied on alpha_EM^INV
    lnlike = lnlike + GetLogLikeFromData(alpha_em_inv, 1d0/NuisP%alphamz)

  end function GetLogLikeNuisance


!**********************************************
! MSSM specific subroutines
!**********************************************

!--------------------------------------------
!  SoftSusy interface
!--------------------------------------------

  subroutine GetSoftSusySpectrum(MSSM, NuisPar, Spect, OutSoft, SoftErr)
    !input:
    !     the parameter values are in  MSSM 
    !     the Nuisance Parameters in NuisPar
    !output:
    !     the SUSY spectrum from SoftSusy in Spect and OutSoft
    !     which is then passed on to DS 
    !     the error flags in SoftErr
  
      Type(Input_params) :: MSSM
      Type(Nuisance_params):: NuisPar
      Type(Susy_spec) :: Spect
      Type(SoftOut) :: OutSoft
      Type(Error_Out) :: SoftErr

    !init softsusy error flags  
      SoftErr%muSqWrongSign = 0
      SoftErr%b= 0
      SoftErr%noRhoConvergence= 0
      SoftErr%tachyon= 0
      SoftErr%nonperturbative= 0
      SoftErr%noMuConvergence= 0
      SoftErr%higgsUfb= 0
      SoftErr%irqfp= 0
      SoftErr%badConvergence= 0   

      
      If (Feedback > 3) write(*,*) '... Now calling softsusy interface ...'

      !GFlags is a gobal variable defined in paramdef.f90
      !it defines what kind of predictions must be computed.
      !Leshouches contains variables to be passed to Feynhiggs
      !also defined in  paramdef.f90

      call softsusy(GFlags, MSSM, NuisPar, Spect, OutSoft, Leshouches, SoftErr)



      If (Feedback > 3) then
         write(*,*) '>>> Softsusy spectrum'
         write(*,'("    Lights Chargino mass   :" G15.7,A)') Spect%mch1, ' (GeV)'
         write(*,'("    Lights Higgs    mass   :" G15.7,A)') Spect%mlh, ' (GeV)'
         write(*,'("    Lights neutralino mass :" G15.7,A)') Spect%mneut1, ' (GeV)'
         write(*,*) '>>> Softsusy: new sparticle masses '
         write(*,'("    selectron R  mass      :" G15.7,A)') Spect%mseR, ' (GeV)'
         write(*,'("    smuon R mass           :" G15.7,A)') Spect%msmuR,  ' (GeV)'  
         write(*,'("    stau 1 mass            :" G15.7,A)') Spect%mstau1, ' (GeV)'   
         write(*,'("    tau sneutrino mass     :" G15.7,A)') Spect%msntau,  ' (GeV)'
         write(*,'("    stop 1 mass            :" G15.7,A)') Spect%mst1, ' (GeV)'
         write(*,'("    sbottom 1 mass         :" G15.7,A)') Spect%msb1, ' (GeV)'
         write(*,'("    Lights squark mass     :" G15.7,A)') min(Spect%mscR,Spect%mssR,Spect%msuR,Spect%msdR), ' (GeV)'
         write(*,'("    Gluino mass            :" G15.7,A)') Spect%mgluino, ' (GeV)'
      end If
    end subroutine GetSoftSusySpectrum

!--------------------------------------------
!  DS and FeynHiggs interface
!--------------------------------------------
    
    subroutine GetPredictions(MSSM, NuisPar, OutSoft, DSO, Prod, BDecay)
      !Input params
      Type(Input_Params) :: MSSM
      Type(Nuisance_params):: NuisPar
      Type(SoftOut) :: OutSoft
      !DS outputs grouped together
      Type(DS_Output) :: DSO
      Type(Prodxbr_Out) :: Prod
      Type(BDecay_Out) :: Bdecay


      if (Feedback > 3) then
         GFlags%Debug = .true.
      else
         GFlags%Debug = .false.
      end if


      if (GFlags%DM_Predict .or. GFlags%DD_Predict .or. GFlags%ID_Predict .or. GFlags%Collider_Predict) then

       if (Feedback > 3) &
         write(*,*) '... Now calling DS interface ...'

       !Flags is a gobal variable defined in paramdef.f90
       !it defines what kind of predictions must be computed.
       if (timing) call StartTiming(Hertz, Begin_Clock)       

       call ds(GFlags, MSSM, NuisPar, OutSoft, DSO)
 
       if (timing) then
         call StopTiming(Begin_Clock, End_Clock, Hertz, Elapsed_Time, ms)
         call ComputeAverage(ds_av,ds_ssq,ms,numav)
       end if

       if (Feedback > 3) call DebugDSOutput(DSO)
 
       if(.not.GFlags%ID%evar .or. (GFlags%ID%evar .and. (GIDin%egam == GIDin%efluxes_i))) then

        ! prints out error flags
        if (Feedback > 2) call DebugDSErrorOutput(DSO%Err)    
       endif


      endif

      if(.not.GFlags%ID%evar .or. (GFlags%ID%evar .and. (GIDin%egam == GIDin%efluxes_i))) then
       if (GFlags%BD_Predict) then
         if (Feedback > 3) write(*,*) '... Now calling BDecay interface...' 
         if (timing) call StartTiming(Hertz,Begin_Clock)       
         call bd(GFlags, MSSM, NuisPar, Outsoft, BDecay, DSO%err)
         if (timing) then
            call StopTiming(Begin_Clock, End_Clock, Hertz, Elapsed_Time, ms)
            call ComputeAverage(bd_av, bd_ssq,ms,numav)
         end if
         ! prints out error flags
         if (Feedback > 2) then 
           write(*,*) '     >>> Error flags from Bdecay (0=OK, 1=not OK)'        
           write(*,'("           BD problem: ", I2)'), DSO%err%bd
         endif
       endif     

       if (GFlags%FH_Predict) then
         if (Feedback > 3) write(*,*) '... Now calling FH interface...' 
         if (timing) call StartTiming(Hertz,Begin_Clock)       
         call FeynHiggs(GFlags, DSO%err, Leshouches, Prod)
         if (timing) then
            call StopTiming(Begin_Clock, End_Clock, Hertz, Elapsed_Time, ms)
            call ComputeAverage(fh_av, fh_ssq,ms,numav)
         end if

         if (Feedback > 3) call DebugFSOutput(Prod)

         ! prints out error flags
         if (Feedback > 2) then
             write(*,*) '   >>> Error flags from FeynHiggs (0=OK, 1=not OK)'        
             write(*,'("         FH problem: ", I2)'), DSO%err%fh
         endif

       endif
      endif
      
    end subroutine GetPredictions

    
!**********************************************
! ANCILLARY SUBROUTINES, output of variables etc
!**********************************************
    subroutine DebugSoftErrorOutput(EOut)
      Type(Error_Out) EOut
      write(*,*) '      >>> Error flags from SoftSusy (0=OK, 1=not OK)'
      write(*,'("            soft no good REWSB   :", I2)'), Eout%muSqWrongSign
      write(*,'("            soft problem with mu :", I2)'), Eout%b
      write(*,'("            soft tachyonic Higgs :", I2)'), Eout%noRhoConvergence
      write(*,'("            soft tachyonic sferm :", I2)'), Eout%tachyon
      write(*,'("            soft Landau pole     :", I2)'), Eout%nonperturbative
      write(*,'("            soft no EWSB         :", I2)'), Eout%noMuConvergence
      write(*,'("            soft no EW minimum   :", I2)'), Eout%higgsUfb
      write(*,'("            soft Landau pole 2   :", I2)'), Eout%irqfp
      write(*,'("            soft no convergence  :", I2)'), Eout%badConvergence


    end subroutine DebugSoftErrorOutput


!**********************************************
! ANCILLARY SUBROUTINES, output of variables etc
!**********************************************
    subroutine DebugDSErrorOutput(EOut)
      Type(Error_Out) EOut

       write(*,*) '     >>> Error flags from DarkSusy (0=OK, 1=not OK)'
       write(*,'("           Neutralino not LSP:", I2)'), Eout%unphys
       write(*,'("           oh2 problem       :", I2)'), Eout%oh2

    end subroutine  DebugDSErrorOutput

    subroutine DebugDSOutput(Out)
      
      Type(DS_Output):: Out

      write(*,*) '>>> DS output'

      if(.not.GFlags%ID%evar .or. (GFlags%ID%evar .and. (GIDin%egam == GIDin%efluxes_i))) then    
       if (GFlags%DM_predict) then
         write(*,*) '     >>> Dark matter properties'
         write(*,'("           Relic density      :", G15.5)') Out%dm%oh2
         write(*,'("           Gaugino fraction   :", G15.5)') Out%dm%gf
       end if
       if (GFlags%DD_predict) then
         write(*,*) '        >> Direct signatures'
         write(*,'("           SI nn cross section:", G15.5,A)') Out%dd%sigsi_nuc, ' (pb)' 
         write(*,'("           SD np cross section:", G15.5,A)') Out%dd%sigsdp, ' (pb)' 
         write(*,'("           SD nn cross section:", G15.5,A)') Out%dd%sigsdn, ' (pb)' 
       end if
      endif
      if (GFlags%ID_predict) then
         write(*,*) '        >> Indirect signatures'
         if(GFlags%ID%gadiff .or. GFlags%ID%gac .or. GFlags%ID%gam &
            .or. GFlags%ID%pb .or. GFlags%ID%db .or. GFlags%ID%ep &
            .or. GFlags%ID%epfrac .or. GFlags%ID%muha &
            .or. GFlags%ID%muhadiff) then
          write (* ,'(A,'//trim(numcat(' ',GIDin%num_hm))//'A)') &
          'Halo models: ', Out%id_in%hmodel(1:GIDin%num_hm)
         endif
         if (GFlags%ID%gadiff) then
          write (* ,'(A,F7.3,A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A)') &
          'Diff gamma ray fluxes with energy ', GIDin%egam, ' GeV: ', Out%id%fluxgadiff(1:GIDin%num_hm), dphigam_units
         endif
         if (GFlags%ID%gac) then
          write (* ,'(A,F7.3,A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A)') &
          'Gamma ray fluxes with threshold enery', GIDin%egath, ' GeV: ', Out%id%fluxgac(1:GIDin%num_hm), phigam_units 
         endif
         if (GFlags%ID%gam) then
          write (* ,'(A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A)') & 
          'Monocromatic gamma ray fluxes gg with energy m_\chi:  ', & 
           Out%id%fluxgaga(1:GIDin%num_hm), phigam_units

          write (* ,'(A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A)') & 
          'Monocromatic gamma ray fluxes gZ with energy m_\chi:  ', & 
           Out%id%fluxgaz(1:GIDin%num_hm), phigam_units 
         endif
         if (GFlags%ID%pb) then        
          write (*,'(A,F7.3,A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A,F7.3)') & 
          'Antiproton fluxes for energy', GIDin%epb, ' GeV: ',&
           Out%id%phipb(1:GIDin%num_hm), ' (GeV^-1 cm^-2 s^-1 sr^-1) &
           for a boost factor: ', GIDin%BF 
         endif
         if (GFlags%ID%db) then          
       write(*,'(A,F7.3,A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A,F7.3)')&
          'Antideutron fluxes for energy', GIDin%edb, ' GeV: ', &
          Out%id%phidb(1:GIDin%num_hm), ' (GeV^-1 cm^-2 s^-1 sr^-1) &
          for a boost factor: ', GIDin%BF 
         endif
         if (GFlags%ID%ep) then
          write(*,'(A,F7.3,A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A,F7.3)')  &
          'Positron fluxes for energy', GIDin%eep, ' GeV: ', &
          Out%id%phiep(1:GIDin%num_hm), ' (GeV^-1 cm^-2 s^-1 sr^-1) for a boost factor: ',  GIDin%BF
         endif
         if (GFlags%ID%epfrac) then          
          write (*,'(A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A,F7.3)') & 
          'Positron fluxes fraction: ', Out%id%phiep_frac(1:GIDin%num_hm), ' for a boost factor: ',  GIDin%BF
         endif
         if (GFlags%ID%musu) then
          write (*,'(A,F7.3,A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A)') &
          'Muon flux from the Sun with threshold energy', GIDin%eth, ' GeV: ',&
           Out%id%ratesu(1:GIDin%num_hm), ratenu_units 
           endif
         if (GFlags%ID%muea) then
          write (*,'(A,F7.3,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A)') &
          'Muon flux from the Earth with threshold energy', GIDin%eth, &
           Out%id%rateea(1:GIDin%num_hm), ratenu_units
         endif
         if (GFlags%ID%musudiff) then
          write(*,'(A,F7.3,A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A)') &
          'Diff muon flux from the Sun with energy', GIDin%enu, ' GeV: ', &
           Out%id%ratesudiff(1:GIDin%num_hm), dratenu_units
         endif
         if (GFlags%ID%mueadiff) then
          write(*,'(A,F7.3,A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A)') &
        'Diff muon flux from the Earth with energy', GIDin%enu, ' GeV: ', &
         Out%id%rateeadiff(1:GIDin%num_hm), dratenu_units
         endif
         if (GFlags%ID%muha) then
          write(*,'(A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A)') & 
          'Muon flux from the Halo: ', Out%id%phimuhalo(1:GIDin%num_hm), ' (km^-2 yr^-1)'       
         endif
         if (GFlags%ID%muhadiff) then
          write(*,'(A,'//trim(numcat(' ',GIDin%num_hm))//'G12.4,A)') & 
          'Diff. muon flux from the Halo: ', Out%id%phimuhalodiff(1:GIDin%num_hm), ' (km^-2 yr^-1 GeV^-1)'       
         endif

         if (GFlags%ID%sigmav) then
          write(*,'("Neutralino annihilation sigma v at p=0: ", G10.4,A)') & 
           Out%id%sigmav, ' (cm^3 s^-1)'
         endif

         if (postproc.and.GFlags%ID%Evar) then
          write(*,'("Energy for fluxes: ", G10.4,A)') & 
           Out%id%E, ' (GeV)'
         endif

      end if
      if(.not.GFlags%ID%evar .or. (GFlags%ID%evar .and. (GIDin%egam == GIDin%efluxes_i))) then
       if (GFlags%Collider_predict) then
         write(*,*) '     >>> Collider quantities'
         write(*,'("           MW               :", G15.7,A)') Out%coll%mw, ' (GeV)'
         write(*,'("           sin2eff          :", G15.5)') Out%coll%sweff2 
         write(*,'("           g-2 amplitude    :", G15.5)') Out%coll%gm2amp
         write(*,'("           b -> s gamma     :", G15.5)') Out%coll%bsgamma
         write(*,'("           Bs -> mu+ mu-    :", G15.5)') Out%coll%brbsmm
         write(*,'("           BRR              :", G15.5)') Out%coll%brr   
         write(*,'("           sgn MSSM/SM      :", I5)')    Out%coll%sgn_MSSMoSM
       end if
     endif

    end subroutine DebugDSOutput


    subroutine DebugFSOutput(prodxbr)

      Type(Prodxbr_Out) :: prodxbr

       write(*,*) '>>> FH output'

       write(*,'(a)')'      >>> Higgs branching ratios and sigma x BR_MSSM/(sigma x BR_SM):'
       write(*,'(a)')

       write(*,'(a)')  '        >> W + h -> h -> b + bbar (TeV):' 
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma (Wh-> h)_MSSM',  'sigma (Wh-> h)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%Wh%sigma_Tev(1), prodxbr%Wh%sigmaSM_Tev(1)
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma x BR(h->bbar)_MSSM',  'sigma BR(h->bbar)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%Wh%bbar_Tev(1), prodxbr%Wh%bbarSM_Tev(1)
       write(*,'(a)')

       write(*,'(5x,a55)')  'sigma x BR(h->bbar)_MSSM/sigma x BR(h->bbar)_SM'
       write(*,'(35x,g11.5)') prodxbr%Wh%bbar_Tev_r(1)
       write(*,'(a)')

       write(*,'(a)')  '        >> Z + h -> h -> b + bbar (TeV):' 
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma (Zh-> h)_MSSM',  'sigma (Zh-> h)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%Zh%sigma_Tev(1), prodxbr%Zh%sigmaSM_Tev(1)
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma x BR(h->bbar)_MSSM',  'sigma BR(h->bbar)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%Zh%bbar_Tev(1), prodxbr%Zh%bbarSM_Tev(1)
       write(*,'(a)')

       write(*,'(5x,a55)')  'sigma x BR(h->bbar)_MSSM/sigma x BR(h->bbar)_SM'
       write(*,'(35x,g11.5)') prodxbr%Zh%bbar_Tev_r(1)
       write(*,'(a)')

       write(*,'(a)')  '        >> bbar + h -> h -> b + bbar (Tev):' 
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma (bbar-> h)_MSSM',  'sigma (bbar-> h)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%bbh%sigma_Tev(1), prodxbr%bbh%sigmaSM_Tev(1)
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma x BR(h->bbar)_MSSM',  'sigma BR(h->bbar)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%bbh%bbar_Tev(1), prodxbr%bbh%bbarSM_Tev(1)
       write(*,'(a)')

       write(*,'(5x,a55)')  'sigma x BR(h->bbar)_MSSM/sigma x BR(h->bbar)_SM'
       write(*,'(35x,g11.5)') prodxbr%bbh%bbar_Tev_r(1)
       write(*,'(a)')

       write(*,'(a)')  '        >> gg -> h -> gamma + gamma (LHC):' 
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma (gg-> h)_MSSM',  'sigma (gg-> h)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%ggh%sigma_LHC(1), prodxbr%ggh%sigmaSM_LHC(1)
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma x BR(h->gamgam)_MSSM',  'sigma BR(h->gamgam)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%ggh%gamgam_LHC(1), prodxbr%ggh%gamgamSM_LHC(1)
       write(*,'(a)')

       write(*,'(5x,a55)')  'sigma x BR(h->gamgam)_MSSM/sigma x BR(h->gamgam)_SM'
       write(*,'(35x,g11.5)') prodxbr%ggh%gamgam_LHC_r(1)
       write(*,'(a)')

       write(*,'(a)')  '        >> qq -> tt h -> b + bar (LHC):' 
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma (qq-> h)_MSSM',  'sigma (qq-> h)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%tth%sigma_LHC(1), prodxbr%tth%sigmaSM_LHC(1)
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma x BR(h->bbar)_MSSM',  'sigma BR(h->bbar)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%tth%bbar_LHC(1), prodxbr%tth%bbarSM_LHC(1)
       write(*,'(a)')

       write(*,'(5x,a55)')  'sigma x BR(h->bbar)_MSSM/sigma x BR(h->bbar)_SM'
       write(*,'(35x,g11.5)') prodxbr%tth%bbar_LHC_r(1)
       write(*,'(a)')

       write(*,'(a)')  '        >> qq -> VV h -> tau + tau (LHC):' 
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma (qq-> VV h)_MSSM',  'sigma (qq-> VV h)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%qqh%sigma_LHC(1), prodxbr%qqh%sigmaSM_LHC(1)
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'sigma x BR(h-> tautau)_MSSM',  'sigma BR(h->VV tautau)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%qqh%tautau_LHC(1), prodxbr%qqh%tautauSM_LHC(1)
       write(*,'(a)')

       write(*,'(5x,a55)')  'sigma x BR(h->tautau)_MSSM/sigma x BR(h->tautau)_SM'
       write(*,'(35x,g11.5)') prodxbr%qqh%tautau_LHC_r(1)
       write(*,'(a)')

       write(*,'(a)')  '        >> Branchings BR(h -> X Y)'
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'BR(h-> b bar)_MSSM',  'BR(h -> b bbar)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%branch%BRtobbar(1) & 
                                    , prodxbr%branch%BRtobbarSM(1)
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'BR(h-> gam gam)_MSSM',  'BR(h -> gam gam)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%branch%BRtogamgam(1) & 
                                   , prodxbr%branch%BRtogamgamSM(1)
       write(*,'(a)')


       write(*,'(2x,a30,4x,a30)')  'BR(h-> tau tau)_MSSM',  'BR(h -> tau tau)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%branch%BRtotautau(1) & 
                                   , prodxbr%branch%BRtotautauSM(1)
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'BR(h-> W W*)_MSSM',  'BR(h -> W W*)_SM'
       write(*,'(18x,g11.5,25x,g11.5)') prodxbr%branch%BRtoWW(1) &
                                  , prodxbr%branch%BRtoWWSM(1)
       write(*,'(a)')

       write(*,'(a)')  '        >> Gamma(h -> X Y)'
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)') 'Gamma(h-> b bbar)_MSSM',  'Gamma(h -> b bbar)_SM'
       write(*,'(15x,g11.5,25x,g11.5)') prodxbr%gamma%tobbar(1) & 
                                   , prodxbr%gamma%tobbarSM(1)
       write(*,'(a)')
       
       write(*,'(2x,a30,4x,a30)')  'Gamma(h-> gam gam)_MSSM',  'Gamma(h -> gam gam)_SM'
       write(*,'(15x,g11.5,25x,g11.5)') prodxbr%gamma%togamgam(1) & 
                                   , prodxbr%gamma%togamgamSM(1)      
       write(*,'(a)')

       write(*,'(2x,a30,4x,a30)')  'Gamma(h-> tau tau)_MSSM',  'Gamma(h -> tau tau)_SM'
       write(*,'(15x,g11.5,25x,g11.5)') prodxbr%gamma%totautau(1) & 
                                   , prodxbr%gamma%totautauSM(1)      
       write(*,'(a)')       

       write(*,'(2x,a30,4x,a30)')  'Gamma(h-> W W*)_MSSM',  'Gamma(h -> W W*)_SM'
       write(*,'(15x,g11.5,25x,g11.5)') prodxbr%gamma%toWW(1) & 
                                   , prodxbr%gamma%toWWSM(1)      
       write(*,'(a)')  




    end subroutine DebugFSOutput



!**********************************************
! NUMERICAL RECIPES SUBROUTINES
!**********************************************
!Integration routines

 SUBROUTINE trapzd(func,a,b,s,n)
   INTEGER n
   REAL*8:: a,b,s,func, af, bf
   EXTERNAL func
   INTEGER it,j
   REAL*8:: del,sum,tnm,x
   if (n.eq.1) then
      
      af = func(a)
      bf = func(b)
      s=0.5*(b-a)*(af+bf)
        !write(*,*) 'funcs', af, bf
      !if all zero, dont try to integrate
      if ((func(a) .eq. 0d0) .and. (func(b) .eq. 0)) return
   else
      it=2**(n-2)
      tnm=it
      del=(b-a)/tnm
      x=a+0.5*del
      sum=0.
      do j=1,it
         sum=sum+func(x)
         x=x+del
      end do
      s=0.5*(s+(b-a)*sum/tnm)
   endif
   return
 END SUBROUTINE trapzd

 !integrates function func from a to b, returns integral in s
 !uses trapzd
 SUBROUTINE qtrap(func,a,b,s)
   INTEGER JMAX
   REAL*8::  a,b,func,s,EPS
   EXTERNAL func
   PARAMETER (EPS=1.e-3, JMAX=100)
   !     USES trapzd
   INTEGER j
   REAL*8::  olds
   olds=-1.e30
   do j=1,JMAX
      !write(*,*) ' j = ', j
      call trapzd(func,a,b,s,j)
      !write(*,*) 's = ', s, abs(s-olds), EPS*abs(olds)
      if (abs(s-olds).le.EPS*abs(olds)) return
      olds=s
   end do
   stop 'too many steps in qtrap'
 end SUBROUTINE qtrap
 
end module CalcLike
