!This module contains 
!        - type definitions for MCMC variables (type definitions for MSSM variables in typedef.f90)
!        - parameter read-in/initialization routines 
!        - mapping routines linking MCMC parameters (params) and the physical parameters
!        - parameters I/O routines
!SuperBayes Package - by Roberto Ruiz de Austri (rruiz@delta.ft.uam.es) and Roberto Trotta (rxt@astro.ox.ac.uk)
!This version October 2007 
!Monte Carlo routines based on COSMOMC package by Antony Lewis (http://cosmologist.info)

module ParamDef

  use parameters !all other types definitions are in there
  use Random
  use settings

 implicit none

 
 Type ParamSet
    real P(num_params)
 end Type ParamSet

 Type ParamScale
    real PMin(num_params), PMax(num_params), PWidth(num_params),center(num_params)
 end Type ParamScale

 integer GridDim(num_params)
 integer :: num_accept=0, numtoget
 integer TotGridPoints

 Type(ParamScale) Scales

 
 logical :: has_propose_matrix = .false.
 logical :: propose_grid = .false., postproc = .false.
 logical :: estimate_propose_matrix = .false.
 logical :: use_log = .false.
 logical :: use_nuisance_splitting = .false. 

 logical :: Use_CDM = .false.
 logical :: Use_LEP = .false.
 logical :: Use_Anomalous_Mu = .false.
 logical :: Use_bsgamma = .false.
 logical :: Use_Bsmumu = .false.
 logical :: Use_DD = .false.
 logical :: Use_ID = .false.
 logical :: Use_Nuisance = .false.
 logical :: Use_Mass_W = .false.
 logical :: Use_Weak_Mixing = .false.
 logical :: Use_Bdecay = .false.
 logical :: CDM_purely_LSP = .true.
 integer :: use_data

 real, dimension(:,:), allocatable :: propose_matrix, propose_matrix_nuis
 real, dimension(:), allocatable :: propose_diag, propose_diag_nuis
 real, dimension(:),   allocatable :: sigmas
 integer, dimension(:), allocatable :: slow_evecs


 !ID variables
 integer :: num_hm, num_ID, num_hmnd

 character(LEN=35) :: phigam_units, dphigam_units, ratenu_units, &
            dratenu_units

 !global variables 
 Type(DS_Output) :: CurrentOutputP, PreviousOutputP
 Type(Susy_spec) :: CurrentSusySp, PreviousSusySp
 Type(SoftOut)   :: CurrentSoftOut, PreviousSoftOut
 Type(Prodxbr_Out) :: CurrentProdxbr, PreviousProdxbr
 Type(Bdecay_Out) :: CurrentBdecay, PreviousBdecay
 Type(Reduced_Out) ::ChainReducedProdxbr

 real :: StartLike = LogZero
 real :: MaxLike = LogZero
 !global variable containing the flags
 Type(Flags) :: GFlags, GFlags_old
 !global variable containing variables to be passed 
 !between softsusy and Feynhiggs
 Type(Leshouches_out) :: Leshouches
 !global variable containing ID params
 Type(ID_in) :: GIDin, GIDin_old 
 !timing variables
 logical, parameter :: timing= .true.
 integer :: Hertz,  Begin_Clock, End_Clock, ms
 character(12) :: Elapsed_Time 
 real :: softsusy_av = 0d0, ds_av = 0d0, fh_av = 0d0, bd_av = 0d0
 real :: like_av = 0d0
 real :: softsusy_ssq = 0d0, ds_ssq = 0d0, fh_ssq = 0d0, bd_ssq = 0d0 
 !standard dev squared
 real :: like_ssq

 !variables needed for the restart
 !used by driver.f90
 logical :: restart, redo_like, redo_theory, redo_change_like_only
 integer :: skip_lines
 character(LEN=5000) RestartLine

 !MPI variables
 double precision    :: MPI_StartTime
contains
           
subroutine Initialize(Params)
        use IniFile
        implicit none
        type (ParamSet) Params
        integer i
        character(LEN=5000) InLine
        character(LEN=120) prop_mat
        real center, wid, mult, like
        real pmat(num_params,num_params)

        
        Ini_fail_on_not_found = .false.
        prop_mat = Ini_Read_String('propose_matrix')
        has_propose_matrix = prop_mat /= ''
 
        if (restart) then
           !RestartLine has been saved before, see driver.f90
           read(RestartLine, *) mult, like, Params%P(1:num_params)
           StartLike = Like
           !need to reconvert alpha_EM
           Params%P(9) = Params%P(9) + alpha_inv_mean
           if (Feedback > 1) then
              write(*,*) ' Restarting from like: ', like
              write(*,*) ' At params ', Params%P(1:num_params)
           end if
        end if
    
        num_params_used = 0
        num_slow = 0
        num_nuis = 0
        TotGridPoints = 1

        do i=1,num_params
   
           InLine = Ini_Read_String(numcat('param',i), .true.)
          
           read(InLine, *, err = 100) center, Scales%PMin(i), Scales%PMax(i), wid, Scales%PWidth(i)
           Scales%center(i) = center
           if (Scales%PMax(i) < Scales%PMin(i)) stop 'You have param Max < Min'
           if (Scales%PWidth(i) /= 0) then
               !those are used params
               num_params_used = num_params_used + 1
              if (i > num_hard) then 
                 num_nuis = num_nuis + 1
              else
                 num_slow = num_slow + 1          
              end if
              if (propose_grid) then
                 GridDim(i) = nint((Scales%PMax(i) - Scales%PMin(i))/Scales%PWidth(i)) 
                 If (Feedback > 0) then 
                    write(*,*) 'Grid Points in Param' ,i, ' = ', GridDim(i)                
                 end If
                 TotGridPoints = TotGridPoints*GridDim(i)
              end if
           end if
           if (.not. restart) then
           
              if (.not. propose_grid) then
                 do
                    if (wid < 0) then
                       !This case we want half gaussian, width -wid
                       !e.g. for positive definite parameters
                       Params%P(i) = center - abs(Gaussian1())*wid
                    else
                       Params%P(i) = center + Gaussian1()*wid
                    end if
                    !Repeat until get acceptable values in range
                    if (Params%P(i)>=  Scales%PMin(i) .and. Params%P(i) <= Scales%PMax(i)) exit
                 end do
              else !if grid
                 Params%P(i) = Scales%PMin(i)
           end if !propose grid
        end if !not restart
        end do

        !write(*,*) 'Original grids', TotGridPoints
        !TotGridPoints is meant to be per chain
        TotGridPoints = nint(1.0*TotGridPoints/(1.0*(nprocs)))

        if (TotGridPoints > numtoget .and. propose_grid) then
           write(*,*) "Grid too large - increase samples to ", TotGridPoints
           stop
        end if
        
        if (propose_grid .and. Feedback > 0) write(*,*) "Number of grid elements per chain: ", TotGridPoints
        allocate(params_used(num_params_used))
        allocate(nuis_params_used(num_nuis))
        num_params_used = 0
        num_nuis = 0
        !now fill in the values
        do i=1,num_params
           if (Scales%PWidth(i) /= 0) then
              num_params_used = num_params_used + 1
              params_used(num_params_used) = i
              if (i > num_hard) then
                num_nuis = num_nuis + 1
                nuis_params_used(num_nuis) = i
             end if
           end if
        end do
   
        ! ---- do you want to use the split slow/nuis or not ----
        ! ---- if not, then num_nuis = 0, num_params_used = num_slow
        if (.not. use_nuisance_splitting) then
           num_nuis = 0
           num_slow = num_params_used
        end if
        if (Feedback > 0 ) then
           write(*, *) 'Varying a total of ', num_params_used, ' parameters'
           if (use_nuisance_splitting) then
              write(*,*)  'of which ', num_slow, ' are SUSY ', num_nuis, ' are nuisance'
              else
              write(*,*)  'all treated as slow parameters / no splitting '
           end if
        end if

        if (has_propose_matrix) then
           
           call ReadMatrix(prop_mat,pmat, num_params, num_params)
           If (Feedback > 0) write(*,*) 'Propose matrix read in'
           !If generated with constrained parameters, assume diagonal in those parameters
           do i=1,num_params
              if (pmat(i,i) ==0 .and. Scales%PWidth(i)/=0) pmat(i,i) = Scales%PWidth(i)**2
           !Enforce new constraints
           if (Scales%PWidth(i)==0) then
              pmat(i,:) = 0
              pmat(:,i) = 0
           end if
           end do
           allocate(propose_matrix(num_params_used, num_params_used))
           !propose_diag is allocated within SetProposeMatrix
           
           propose_matrix = pmat(params_used, params_used)
           call SetProposeMatrix
           if (Feedback > 0) then 
              write(*,*) 'which are the params_used', params_used 
              write(*,*) 'which are the nuis_params_used', nuis_params_used
           end if
           !look for it in this file below
           !on output, propose_matrix contains the EVectors
           !propose_diag contains the diagonal EValues
        end if
   

        return
100     write(*,*) 'Error reading param details: '//trim(InLIne)
        stop

end subroutine Initialize

subroutine ParamsToMSSMParams(Params, MSSM, NuisP)
     use settings

     implicit none
     real Params(num_Params)
     Type(Input_Params) MSSM
     Type(Nuisance_Params) NuisP

     if (use_log) then
        MSSM%m0 = 10**(Params(1))
        MSSM%mhalf = 10**(Params(2))
     else
        MSSM%m0 = abs(Params(1))
        MSSM%mhalf = abs(Params(2)) !abs used for reflection strategy
     end if
     MSSM%a0 = Params(3)
     MSSM%tanb = Params(4)
     MSSM%sgnmu0 = Params(5)
     !this is fixed for the moment
     MSSM%scaledep = 1.0d0 
     !nuisance parameters
     NuisP%mbottom = Params(6)
     NuisP%mtop = Params(7)
     NuisP%alphasmz = Params(8)
     !param(9) is alpha_EM^-1
     NuisP%alphamz = 1d0/Params(9)
     !those are fixed for the moment
     NuisP%ftp_s = 0.118 
     NuisP%ftn_s = 0.118

     if (Feedback > 3)  call PrintOutModelParams(MSSM, NuisP)
 
end subroutine ParamsToMSSMParams

subroutine MSSMParamsToParams(MSSM, Params, NuisP)
     use settings

     implicit none
     real Params(num_Params)
     Type(Input_Params) MSSM
     Type(Nuisance_Params) NuisP

     if (use_log) then
        Params(1) = LOG10(MSSM%m0) 
        Params(2) = LOG10(MSSM%mhalf)
     else
        Params(1) = MSSM%m0
        Params(2) = MSSM%mhalf
     end if
     Params(3) = MSSM%a0 
     Params(4) = MSSM%tanb
     Params(5) = MSSM%sgnmu0

     !nuisance parameters
     Params(6) = NuisP%mbottom
     Params(7) = NuisP%mtop
     Params(8) = NuisP%alphasmz
     !param(9) is alpha_EM^-1
     Params(9) = 1d0/NuisP%alphamz
     !this is fixed for the moment - in case you forgot to set it!
     MSSM%scaledep = 1.0d0 
     NuisP%ftp_s = 0.118 
     NuisP%ftn_s = 0.118

end subroutine MSSMParamsToParams

subroutine PrintOutModelParams(InP, InN)
     Type(Input_Params) InP
     Type(Nuisance_Params) InN

        write(*,*) '>>>> Input Params'
        write(*,*) 'm0  param    : ', InP%m0, ' (GeV)' 
        write(*,*) 'mhalf  param : ', InP%mhalf, ' (GeV)'
        write(*,*) 'a0 param     : ', InP%a0, ' (GeV)'
        write(*,*) 'tanb param   : ', InP%tanb
        write(*,*) 'sgnmu0 param : ', InP%sgnmu0
        write(*,*) '>>>> Nuisance Params'
        write(*,*) 'mb(mb)^MSbar : ', InN%mbottom, ' (GeV)' 
        write(*,*) 'm top        : ', InN%mtop, ' (GeV)'
        write(*,*) 'alpha_s^MSbar : ', InN%alphasmz
        write(*,*) 'alpha_em^MSbar :', InN%alphamz
        write(*,*) '  or al_em^-1: ', 1.0d0/InN%alphamz
        write(*,*) '>>>> Fixed Nuisance Params'
        write(*,*) 'ftp_s        : ', InN%ftp_s
        write(*,*) 'ftp_n        : ', InN%ftn_s

end subroutine PrintOutModelParams


subroutine WriteParams(P, mult, like)
    implicit none
    Type(ParamSet) P
    real, intent(in) :: mult, like
    real(8) :: Xt, Xb, mu, tan_2thetat, tan_2thetab
    Type(Reduced_Out) :: ReducedProdxbr

    if (outfile_unit == 0) return

    !quantities which have not been computed should simply be zero
    !as they have been initialized
    !for reflection strategy, takes the abs of the masses
    P%P(1) = abs(P%P(1))
    P%P(2) = abs(P%P(2))


    if(.not.GFlags%ID%evar .or. (GFlags%ID%evar .and. (GIDin%egam == GIDin%efluxes_i))) call ConvertForWriting(P)

    
    if(GFlags%FH_predict) then
      !in ReducedProdxbr I only keep relevant output to reduce output size
      call ReduceOutput(PreviousProdxbr, ReducedProdxbr)
      if ((postproc) .and. (.not. redo_theory)) ReducedProdxbr = ChainReducedProdxbr
      !dealing with special case for postprocessing
    endif

    !Compute Xt, Xb quantities for output
    !Add here computation of other quantities if you want
    mu = PreviousSoftOut%soft%mu 
    Xt = PreviousSoftOut%soft%asoftu(3) -  mu/P%P(4)
    Xb = PreviousSoftOut%soft%asoftd(3) -  mu*P%P(4)
    !computes stop and sbottom mixing angles for output
    tan_2thetat = tan(2.0*PreviousSoftOut%sfer%thetat)
    tan_2thetab = tan(2.0*PreviousSoftOut%sfer%thetab) 

 
    !Which variables to save
    !Customize it to suits your need
    !Remember to adjust the formats in SetFormat
    write (outfile_unit, trim(fmt_params), ADVANCE='NO') mult,like, P%P
    if (GFlags%DM_predict)  write (outfile_unit, fmt_dm, ADVANCE='NO') PreviousOutputP%dm
    if (GFlags%DD_predict)  write (outfile_unit, fmt_dd, ADVANCE='NO') PreviousOutputP%dd
    if (GFlags%Collider_predict) write (outfile_unit, fmt_coll, ADVANCE='NO') PreviousOutputP%coll
    if (GFlags%BD_predict)  write (outfile_unit, fmt_bdecay, ADVANCE='NO') PreviousBdecay
    write (outfile_unit, fmt_susysp, ADVANCE='NO')  PreviousSusySp
    write (outfile_unit, fmt_xtxbmu, ADVANCE='NO')  Xt, Xb, mu
    write (outfile_unit, fmt_tantheta, ADVANCE='NO')  tan_2thetat, tan_2thetab
    if (GFlags%FH_predict)  write (outfile_unit, fmt_prodxbr, ADVANCE='NO') ReducedProdxbr
!
    if (GFlags%ID_predict)  then
      if (GFlags%ID%gadiff) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%fluxgadiff(1:num_hm)
      if (GFlags%ID%gac) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%fluxgac(1:num_hm)
      if (GFlags%ID%gam) then
       write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%fluxgaga(1:num_hm)    
       write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%fluxgaz(1:num_hm)
      endif
      if (GFlags%ID%pb) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%phipb (1:num_hm)
      if (GFlags%ID%db) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%phidb(1:num_hm)
      if (GFlags%ID%ep) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%phiep(1:num_hm)
      if (GFlags%ID%epfrac) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%phiep_frac(1:num_hm)
      if (GFlags%ID%musu) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%ratesu(1:num_hm)
      if (GFlags%ID%muea) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%rateea(1:num_hm)
      if (GFlags%ID%musudiff) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%ratesudiff(1:num_hm)
      if (GFlags%ID%mueadiff) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%rateeadiff(1:num_hm)
      if (GFlags%ID%muha) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%phimuhalo(1:num_hm) 
      if (GFlags%ID%muhadiff) write (outfile_unit, fmt_idhmd, ADVANCE='NO') PreviousOutputP%id%phimuhalodiff(1:num_hm)
      if (GFlags%ID%sigmav) write (outfile_unit, fmt_idhmid, ADVANCE='NO') PreviousOutputP%id%sigmav
      if (postproc.and.GFlags%ID%evar) write (outfile_unit, fmt_idhmid, ADVANCE='NO') PreviousOutputP%id%e
    endif

    
    write (outfile_unit, '(A)') ' ' !to start a new line
    
    if (flush_write) call FlushFile(outfile_unit)

end  subroutine WriteParams

subroutine SetFormat
  character(len=3) :: nstr
  integer :: i,n
  Type(Reduced_Out) :: TmpProdxbr
  character(LEN=lmax), dimension(5):: MixingNames= (/ '\mu (GeV)', 'X_t (GeV)', 'X_b (GeV)', 'tan(\theta_t)', 'tan(theta_b)'/)


  !The variables are defined in settings.f90
  fmt_params =  '('//trim(numcat('2E15.5,',num_params))//'E20.10)'
  fmt_dm = '(2E15.5)'
  fmt_dd = '(3E15.5)'
  fmt_coll = '(4E15.5,I3,2E15.5)' 
  fmt_susysp = '(33E16.6)'
  fmt_bdecay = '(1E15.5)'
  fmt_xtxbmu = '(3E15.5)'
  fmt_tantheta = '(2E15.5)'
  !Format for FH output 
  fmt_prodxbr = '(46E15.5)'
  !Format for ID output
  fmt_idhmd = '('//trim(numcat(' ',num_hm))//'E15.5)'
  fmt_idhmid = '(E15.5)'



  if(GIDin%delta_gamma == 0.d0) then 
    phigam_units = ' (cm^-2 s^-1 sr^-1)'
    dphigam_units = ' (cm^-2 s^-1 GeV^-1 sr^-1)' 
  else
    phigam_units = ' (cm^-2 s^-1)'
    dphigam_units = ' (cm^-2 s^-1 GeV^-1)'   
  endif

  if(GIDin%rtype == 2) then
    ratenu_units =  ' (km^-3 yr^-1)'
    dratenu_units =  ' (km^-3 yr^-1 GeV^-1 degree^-1)'
  else
    ratenu_units =  ' (km^-2 yr^-1)'
    dratenu_units =  ' (km^-2 yr^-1 GeV^-1 degree^-1)'
  endif
   


  n = 0
  !writes .info file
  ParamsNames(1:5) = pack(CMSSMNames,CMSSMUsed)
  ParamsNames(6:9) = pack(NuisanceNames,NuisanceUsed)
  write(infofile_unit, '(A)') '#first col contains the multiplicity'
  write(infofile_unit, '(A)') '#second col contains -2ln(like)=chisq'
  write(infofile_unit, '(A)') '#subsequent columns: col i+2 contain parameter with name lab i'
  write(infofile_unit, '(A)') '### input and nuisance params'    
  do i=1,num_params
     n = n+1
     write(nstr,'(I3)') n
     write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(ParamsNames(i))
  end do
  if (GFlags%DM_predict) then
     write(infofile_unit, '(A)') '### Dark matter parameters'
     do i=1,size(DMNames)
        n = n+1
        write(nstr,'(I3)') n
        write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(DMNames(i))
     end do
  end if
  
  if (GFlags%DD_predict) then
     write(infofile_unit, '(A)') '### Direct Detection parameters'
     do i=1,size(DDNames)
        n = n+1
        write(nstr,'(I3)') n
        write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(DDNames(i))
     end do
  end if
  
  if (GFlags%Collider_predict) then
     write(infofile_unit, '(A)') '### Collider quantities'
     do i=1,size(ColliderNames)
        n = n+1
        write(nstr,'(I3)') n
        write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(ColliderNames(i))
     end do
  end if
  
  if (GFlags%BD_predict) then
     write(infofile_unit, '(A)') '### B decay quantities'
     do i=1,size(BDNames)
        n = n+1
        write(nstr,'(I3)') n
        write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(BDNames(i))
     end do
  end if
  
  write(infofile_unit, '(A)') '### SUSY spectrum'
  do i=1,size(SusySpecNames)
     n = n+1
     write(nstr,'(I3)') n
     write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(SusySpecNames(i))
  end do
  
  write(infofile_unit, '(A)') '### Mixing parameters'
  do i=1,size(MixingNames)
     n = n+1
     write(nstr,'(I3)') n
     write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(MixingNames(i))
  end do 
 
 if (GFlags%FH_predict) then
    !Only needed to set names correctly.
    call ReduceOutput(PreviousProdxbr, TmpProdxbr)
    write(infofile_unit, '(A)') '### BRs and cross sections'
    do i=1,size(FHReducedNames)
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(FHReducedNames(i))
    end do

 end if


 if (GFlags%ID_predict) then
    write(infofile_unit, '(A)') '### Indirect Detection parameters'

    if (GFlags%ID%gadiff) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(1))//trim(dphigam_units)//' ['//trim(GIDin%hmodel(i))//']'
     enddo
    endif

    if (GFlags%ID%gac) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(2))//trim(dphigam_units)//' ['//trim(GIDin%hmodel(i))//']'
     enddo
    endif

    if (GFlags%ID%gam) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(3))//trim(dphigam_units)//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif

    if (GFlags%ID%gam) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(4))//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif

    if (GFlags%ID%pb) then 
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(5))//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif

    if (GFlags%ID%db) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(6))//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif


    if (GFlags%ID%ep) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(7))//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif

    if (GFlags%ID%epfrac) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(8))//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif

    if (GFlags%ID%musu) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(9))//trim(ratenu_units)//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif


    if (GFlags%ID%muea) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(10))//trim(ratenu_units)//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif

    if (GFlags%ID%musudiff) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(11))//trim(dratenu_units)//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif


    if (GFlags%ID%mueadiff) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(12))//trim(dratenu_units)//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif


    if (GFlags%ID%muha) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(13))//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif

    if (GFlags%ID%muhadiff) then
     do i=1,num_hm
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(14))//' ['//trim(GIDin%hmodel(i))//']'
     end do
    endif


    if (GFlags%ID%sigmav) then
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(15))
    endif

    if (postproc.and.GFlags%ID%Evar) then
       n = n+1
       write(nstr,'(I3)') n
       write(infofile_unit, '(4A)') 'lab'//trim(adjustl(nstr))//'= '//trim(IDNames(16))
    endif



 end if


end subroutine SetFormat


subroutine ConvertForWriting(P)

    Type(ParamSet) P
    !for alpha_em_INV write only the difference wrt mean value
    !to reduce number of necessary digits
    !the constants are defined in settings.f90 and are NOT 
    !updated when exp data change, in order to mantain
    !backwards compatibility with previous chains    
     P%P(9) = P%P(9) - alpha_inv_mean
    !the same for sineff^2 - this is a global variable anyways
    !the same for M_W
    PreviousOutputP%coll%sweff2 = PreviousOutputP%coll%sweff2 - sw2eff_mean
    PreviousOutputP%coll%mw = PreviousOutputP%coll%mw -  mW_mean

end subroutine ConvertForWriting

subroutine ConvertFromReading(P, DS)

    Type(ParamSet) P
    Type(DS_Output), optional :: DS


    P%P(9) = P%P(9) +  alpha_inv_mean
   

    if (present(DS)) then
       DS%coll%sweff2 =  DS%coll%sweff2 + sw2eff_mean 
       DS%coll%mw  = DS%coll%mw + mW_mean
    end if

end subroutine ConvertFromReading

subroutine ReduceOutput(Prodxbr, Reduced)

    type(Prodxbr_Out):: Prodxbr
    type(Reduced_Out):: Reduced
    !keep the info you want to save in the chain files
    !W + h -> h -> b + bbar (TeV)
    Reduced%Wh_sigma_Tev = Prodxbr%Wh%sigma_Tev(1)
    Reduced%Wh_sigmaSM_Tev = Prodxbr%Wh%sigmaSM_Tev(1)
    Reduced%Wh_bbar_Tev =  Prodxbr%Wh%bbar_Tev(1)
    Reduced%Wh_bbarSM_Tev =  Prodxbr%Wh%bbarSM_Tev(1)
    FHReducedNames(1:4) =  FHWhNames(1:4)
    !Z + h -> h -> b + bbar (TeV) 
    Reduced%Zh_sigma_Tev = Prodxbr%Zh%sigma_Tev(1)
    Reduced%Zh_sigmaSM_Tev =  Prodxbr%Zh%sigmaSM_Tev(1)
    Reduced%Zh_bbar_Tev = Prodxbr%Zh%bbar_Tev(1) 
    Reduced%Zh_bbarSM_Tev = Prodxbr%Zh%bbarSM_Tev(1)
    FHReducedNames(5:8) =  FHZhNames(1:4)
    !bbar + h -> h -> b + bbar (Tev)
    Reduced%bbh_sigma_Tev = Prodxbr%bbh%sigma_Tev(1)
    Reduced%bbh_sigmaSM_Tev = Prodxbr%bbh%sigmaSM_Tev(1)
    Reduced%bbh_bbar_Tev = Prodxbr%bbh%bbar_Tev(1)
    Reduced%bbh_bbarSM_Tev = Prodxbr%bbh%bbarSM_Tev(1)
    FHReducedNames(9:12) =  FHbbhNames(1:4)
    !gg -> h -> gamma + gamma (LHC)
    Reduced%ggh_sigma_Tev = Prodxbr%ggh%sigma_Tev(1)
    Reduced%ggh_sigmaSM_Tev = Prodxbr%ggh%sigmaSM_Tev(1)
    Reduced%ggh_sigma_LHC = Prodxbr%ggh%sigma_LHC(1)
    Reduced%ggh_sigmaSM_LHC = Prodxbr%ggh%sigmaSM_LHC(1)
    Reduced%ggh_gamgam_LHC = Prodxbr%ggh%gamgam_LHC(1)
    Reduced%ggh_gamgamSM_LHC = Prodxbr%ggh%gamgamSM_LHC(1)
    FHReducedNames(13:18) =  FHgghNames(1:6)
    !qq -> tt h -> b + bar (LHC)
    Reduced%tth_sigma_LHC = Prodxbr%tth%sigma_LHC(1)
    Reduced%tth_sigmaSM_LHC = Prodxbr%tth%sigmaSM_LHC(1)
    Reduced%tth_bbar_LHC = Prodxbr%tth%bbar_LHC(1)
    Reduced%tth_bbarSM_LHC = Prodxbr%tth%bbarSM_LHC(1)
    FHReducedNames(19:22) =  FHtthNames(1:4)
    !qq -> VV h -> tau + tau (LHC)
    Reduced%qqh_sigma_LHC = Prodxbr%qqh%sigma_LHC(1)
    Reduced%qqh_sigmaSM_LHC = Prodxbr%qqh%sigmaSM_LHC(1)
    Reduced%qqh_tautau_LHC = Prodxbr%qqh%tautau_LHC(1)
    Reduced%qqh_tautauSM_LHC = Prodxbr%qqh%tautauSM_LHC(1)
    FHReducedNames(23:26) =  FHqqhNames(1:4)
    !Branchings Reduced% BR(h -> X Y)
    !BR(h-> b bar)_MSSM & SM, BR(h-> gam gam)_MSSM & SM,BR(h-> tau tau)_MSSM & SM
    !new channel added in Sept 06
    Reduced%bh_sigma_Tev = Prodxbr%bh%sigma_Tev(1)
    Reduced%bh_sigmaSM_Tev = Prodxbr%bh%sigmaSM_Tev(1)
    Reduced%bh_sigma_LHC = Prodxbr%bh%sigma_LHC(1)
    Reduced%bh_sigmaSM_LHC = Prodxbr%bh%sigmaSM_LHC(1)
    FHReducedNames(27:30) =  FHbhNames(1:12:3)

    Reduced%branch_BRtobbar = Prodxbr%branch%BRtobbar(1)
    Reduced%branch_BRtobbarSM = Prodxbr%branch%BRtobbarSM(1)
    Reduced%branch_BRtogamgam = Prodxbr%branch%BRtogamgam(1)
    Reduced%branch_BRtogamgamSM = Prodxbr%branch%BRtogamgamSM(1)
    Reduced%branch_BRtotautau = Prodxbr%branch%BRtotautau(1)
    Reduced%branch_BRtotautauSM = Prodxbr%branch%BRtotautauSM(1)
    !new channel added in June 06
    Reduced%branch_BRtoWW = Prodxbr%branch%BRtoWW(1)
    Reduced%branch_BRtoWWSM = Prodxbr%branch%BRtoWWSM(1)
    FHReducedNames(31:38) =  FHBRNames(1:24:3)
    !Widths
    Reduced%gamma_tobbar = Prodxbr%gamma%tobbar(1)
    Reduced%gamma_tobbarSM = Prodxbr%gamma%tobbarSM(1)
    Reduced%gamma_togamgam = Prodxbr%gamma%togamgam(1)
    Reduced%gamma_togamgamSM = Prodxbr%gamma%togamgamSM(1)
    Reduced%gamma_totautau = Prodxbr%gamma%totautau(1)
    Reduced%gamma_totautauSM = Prodxbr%gamma%totautauSM(1)
    Reduced%gamma_toWW = Prodxbr%gamma%toWW(1)
    Reduced%gamma_toWWSM = Prodxbr%gamma%toWWSM(1)
    FHReducedNames(39:46) =  FHwidthNames(1:24:3) 
end subroutine ReduceOutput


subroutine WriteDebug(Flag, InP, InN, Spect, DSO, Prodxbr, BDecay, SoftO)
 
    Type(FLAGS), INTENT(IN) :: Flag
    Type(Input_params), INTENT(IN) :: InP
    Type(Nuisance_params), INTENT(IN) :: InN
    Type(Susy_spec), INTENT(IN) :: Spect
    Type(DS_Output), INTENT(IN) :: DSO
    Type(Prodxbr_Out), INTENT(IN) :: Prodxbr
    Type(BDecay_Out), INTENT(IN) :: BDecay
    Type(SoftOut), INTENT(IN) :: SoftO

    integer, parameter ::  nout=99
    integer :: n_hm 
    real :: Xb, Xt
    character(LEN = 200) :: fmt_idhm, fmt_idhmd, fmt_idhmnd 


    n_hm = GIDin%num_hm 

    fmt_idhm = '(5x,'//trim(numcat(' ',n_hm))//'A15)'
    fmt_idhmd =  '('//trim(numcat(' ',n_hm))//'E15.5)'
    fmt_idhmnd = '(E15.5)'


    if(GIDin%delta_gamma == 0.d0) then 
     phigam_units = ' (cm^-2 s^-1 sr^-1)'
     dphigam_units = ' (cm^-2 s^-1 GeV^-1 sr^-1)' 
    else
     phigam_units = ' (cm^-2 s^-1)'
     dphigam_units = ' (cm^-2 s^-1 GeV^-1)'   
    endif

    if(GIDin%rtype == 2) then
     ratenu_units =  ' (km^-3 yr^-1)'
     dratenu_units =  ' (km^-3 yr^-1 GeV^-1 degree^-1)'
    else
     ratenu_units =  ' (km^-2 yr^-1)'
     dratenu_units =  ' (km^-2 yr^-1 GeV^-1 degree^-1)'
    endif


    open(nout,file='tester.out',status='replace')


    write(nout,'(a)')'                  OUTPUT: CMSSM CASE'    
    write(nout,'(a)')'                  -------------------'
    write(nout,'(a)')
    !
    write(nout,'(a)')'Input values:'
    write(nout,'(a)')'-------------'
    write(nout,578)'m_0','m_1/2','A_0','tan(beta)','sign(mu)'
    write(nout,100) InP%m0, InP%mhalf, InP%a0 , InP%tanb, InP%sgnmu0
    write(nout,'(a)') 
    !
    write(nout,581)'M_top','mb_mb','1/alpha','alpha_S' 
    write(nout,102) InN%mtop, InN%mbottom ,InN%alphamz, InN%alphasmz
    write(nout,'(a)') 
    !
    write(nout,'(a)')'ElectroWeak parameters:'
    write(nout,'(a)')'-----------------------'     
    !
    write(nout,583)'mW  ','sin^2 theta_eff'
    write(nout,110)  DSO%Coll%mW, DSO%Coll%sweff2
    write(nout,'(a)')  
    !
    write(nout,'(a)')'Higgs and SUSY particle masses (GeV): '
    write(nout,'(a)')'-------------------------------------'
    !
    write(nout,585)'h ','H','A','H+'
    write(nout,111) Spect%mlh, Spect%mhh, Spect%mA, Spect%mhc
    write(nout,'(a)')
    !
    write(nout,587)'chi+_1','chi+_2','chi0_1','chi0_2','chi0_3', 'chi0_4' 
    write(nout,104) Spect%mch1,Spect%mch2,Spect%mneut1,Spect%mneut2 &
         ,Spect%mneut3 ,Spect%mneut4
    write(nout,'(a)')
    !
    write(nout,590)'gluino'
    write(nout,106) Spect%mgluino
    write(nout,'(a)')
    !
    write(nout,591)'stop_1','stop_2','sup_R','sup_L','scharm_R','scharm_L'
    write(nout,101) Spect%mst1, Spect%mst2, Spect%msuR, Spect%msuL &
         ,Spect%mscR, Spect%mscL
    write(nout,'(a)')
    
    write(nout,592)'sbot_1','sbot_2','sdown_R','sdown_L', 'sstrang_R' &
         ,'sstrang_L' 
    write(nout,101)  Spect%msb1,  Spect%msb2, Spect%msdR, Spect%msdL &
         , Spect%mssR, Spect%mssL         
    write(nout,'(a)')
    
    write(nout,593)'stau_1','stau_2','snutau'
    write(nout,101)  Spect%mstau1,  Spect%mstau2, Spect%msntau
    write(nout,'(a)')
    
    write(nout,594)'selec_R','selec_L','snuelec'
    write(nout,103) Spect%mseR, Spect%mseL, Spect%msne
    write(nout,'(a)')
    
    write(nout,594)'smuon_R','smuon_L','snumuon'
    write(nout,103) Spect%msmuR, Spect%msmuL, Spect%msnmu
    write(nout,'(a)') 

    !
    if (Flag%DM_predict) then 
       write(nout,'(a)')'Relic density:'
       write(nout,'(a)')'--------------'
       
       write(nout,112)'oh2','gaugino fraction'
       write(nout,310) DSO%DM%oh2, DSO%DM%gf
       write(nout,'(a)') 
    endif
    !
    if (Flag%DD_predict) then
       write(nout,'(a)')'Direct detection cross sections:'
       write(nout,'(a)')'--------------------------------'
       write(nout,113) 'sigsi_nuc (pb)','sigsdp (pb)','sigsdn (pb)'
       write(nout,311) DSO%DD%sigsi_nuc, DSO%DD%sigsdp,DSO%DD%sigsdn
       write(nout,'(a)')
    endif
    !
    if (Flag%Collider_predict) then
       write(nout,'(a)')'Collider Observables:'
       write(nout,'(a)')'---------------------'

       write(nout,114)'b->s gamma','sgn_MSSMoSM','b->m+ mu-','a_mu','D_MB (ps^-1)'
       write(nout,312) DSO%Coll%bsgamma, DSO%Coll%sgn_MSSMoSM &
            , DSO%Coll%brbsmm, DSO%Coll%gm2amp,BDecay%delMB
       write(nout,'(a)') 
    endif
    !
    write(nout,'(a)')'Mixing parameters (GeV):'
    write(nout,'(a)')'------------------'
    write(nout,691)'asoftu(3)','asoftd(3)','mu', 'X_t','X_b'
    Xt =  SoftO%soft%asoftu(3) -  SoftO%soft%mu/InP%tanb
    Xb =  SoftO%soft%asoftd(3) -  SoftO%soft%mu*InP%tanb
    write(nout,692) SoftO%soft%asoftu(3),  SoftO%soft%asoftd(3),  SoftO%soft%mu, Xt, Xb
    write(nout,'(a)') 
    !
    if (Flag%FH_predict) then
       write(nout,'(a)')'Higgs branching ratios and sigma x BR_MSSM/(sigma x BR_SM):'
       write(nout,'(a)')'-----------------------------------------------------------'
       write(nout,'(a)')

       write(nout,'(a)')  'W + h -> h -> b + bbar (TeV):' 
       write(nout,'(a)')'------------------------------'

       write(nout,115)  'sigma (Wh-> h)_MSSM',  'sigma (Wh-> h)_SM'
       write(nout,313) prodxbr%Wh%sigma_Tev(1), prodxbr%Wh%sigmaSM_Tev(1)
       write(nout,'(a)')

       write(nout,119)  'sigma x BR(h->bbar)_MSSM',  'sigma BR(h->bbar)_SM'
       write(nout,313) prodxbr%Wh%bbar_Tev(1), prodxbr%Wh%bbarSM_Tev(1)
       write(nout,'(a)')

       write(nout,120)  'sigma x BR(h->bbar)_MSSM/sigma x BR(h->bbar)_SM'
       write(nout,317) prodxbr%Wh%bbar_Tev_r(1)
       write(nout,'(a)')

       write(nout,'(a)')  'Z + h -> h -> b + bbar (TeV):' 
       write(nout,'(a)')'------------------------------'

       write(nout,115)  'sigma (Zh-> h)_MSSM',  'sigma (Zh-> h)_SM'
       write(nout,313) prodxbr%Zh%sigma_Tev(1), prodxbr%Zh%sigmaSM_Tev(1)
       write(nout,'(a)')

       write(nout,119)  'sigma x BR(h->bbar)_MSSM',  'sigma BR(h->bbar)_SM'
       write(nout,313) prodxbr%Zh%bbar_Tev(1), prodxbr%Zh%bbarSM_Tev(1)
       write(nout,'(a)')

       write(nout,120)  'sigma x BR(h->bbar)_MSSM/sigma x BR(h->bbar)_SM'
       write(nout,317) prodxbr%Zh%bbar_Tev_r(1)
       write(nout,'(a)')

       write(nout,'(a)')  'bbar + h -> h -> b + bbar (Tev):' 
       write(nout,'(a)')'----------------------------------'

       write(nout,115)  'sigma (bbar-> h)_MSSM',  'sigma (bbar-> h)_SM'
       write(nout,313) prodxbr%bbh%sigma_Tev(1), prodxbr%bbh%sigmaSM_Tev(1)
       write(nout,'(a)')

       write(nout,119)  'sigma x BR(h->bbar)_MSSM',  'sigma BR(h->bbar)_SM'
       write(nout,313) prodxbr%bbh%bbar_Tev(1), prodxbr%bbh%bbarSM_Tev(1)
       write(nout,'(a)')

       write(nout,120)  'sigma x BR(h->bbar)_MSSM/sigma x BR(h->bbar)_SM'
       write(nout,317) prodxbr%bbh%bbar_Tev_r(1)
       write(nout,'(a)')

       write(nout,'(a)')  'gg -> h -> gamma + gamma (LHC):' 
       write(nout,'(a)')'----------------------------------'

       write(nout,115)  'sigma (gg-> h)_MSSM',  'sigma (gg-> h)_SM'
       write(nout,313) prodxbr%ggh%sigma_LHC(1), prodxbr%ggh%sigmaSM_LHC(1)
       write(nout,'(a)')

       write(nout,119)  'sigma x BR(h->gamgam)_MSSM',  'sigma BR(h->gamgam)_SM'
       write(nout,313) prodxbr%ggh%gamgam_LHC(1), prodxbr%ggh%gamgamSM_LHC(1)
       write(nout,'(a)')

       write(nout,120)  'sigma x BR(h->gamgam)_MSSM/sigma x BR(h->gamgam)_SM'
       write(nout,317) prodxbr%ggh%gamgam_LHC_r(1)
       write(nout,'(a)')

       write(nout,'(a)')  'qq -> tt h -> b + bar (LHC):' 
       write(nout,'(a)')'----------------------------------'

       write(nout,115)  'sigma (qq-> h)_MSSM',  'sigma (qq-> h)_SM'
       write(nout,313) prodxbr%tth%sigma_LHC(1), prodxbr%tth%sigmaSM_LHC(1)
       write(nout,'(a)')

       write(nout,119)  'sigma x BR(h->bbar)_MSSM',  'sigma BR(h->bbar)_SM'
       write(nout,313) prodxbr%tth%bbar_LHC(1), prodxbr%tth%bbarSM_LHC(1)
       write(nout,'(a)')

       write(nout,120)  'sigma x BR(h->bbar)_MSSM/sigma x BR(h->bbar)_SM'
       write(nout,317) prodxbr%tth%bbar_LHC_r(1)
       write(nout,'(a)')

       write(nout,'(a)')  'qq -> VV h -> tau + tau (LHC):' 
       write(nout,'(a)')'--------------------------------'

       write(nout,115)  'sigma (qq-> VV h)_MSSM',  'sigma (qq-> VV h)_SM'
       write(nout,313) prodxbr%qqh%sigma_LHC(1), prodxbr%qqh%sigmaSM_LHC(1)
       write(nout,'(a)')

       write(nout,119)  'sigma x BR(h-> tautau)_MSSM',  'sigma BR(h->VV tautau)_SM'
       write(nout,313) prodxbr%qqh%tautau_LHC(1), prodxbr%qqh%tautauSM_LHC(1)
       write(nout,'(a)')

       write(nout,120)  'sigma x BR(h->tautau)_MSSM/sigma x BR(h->tautau)_SM'
       write(nout,317) prodxbr%qqh%tautau_LHC_r(1)
       write(nout,'(a)')

       write(nout,'(a)')  'Branchings BR(h -> X Y)'
       write(nout,'(a)')'-------------------------'

       write(nout,115)  'BR(h-> b bbar)_MSSM',  'BR(h -> b bbar)_SM'
       write(nout,313) prodxbr%branch%BRtobbar(1) & 
                                    , prodxbr%branch%BRtobbarSM(1)
       write(nout,'(a)')

       write(nout,115)  'BR(h-> gam gam)_MSSM',  'BR(h -> gam gam)_SM'
       write(nout,313) prodxbr%branch%BRtogamgam(1) & 
                                   , prodxbr%branch%BRtogamgamSM(1)
       write(nout,'(a)')


       write(nout,115)  'BR(h-> tau tau)_MSSM',  'BR(h -> tau tau)_SM'
       write(nout,313) prodxbr%branch%BRtotautau(1) & 
                                   , prodxbr%branch%BRtotautauSM(1)
       write(nout,'(a)')

       write(nout,115)  'BR(h-> W W*)_MSSM',  'BR(h -> W W*)_SM'
       write(nout,313) prodxbr%branch%BRtoWW(1) &
                                  , prodxbr%branch%BRtoWWSM(1)
       write(nout,'(a)')

       write(nout,'(a)')  'Gamma(h -> X Y)'
       write(nout,'(a)')'-----------------'

       write(nout,115)  'Gamma(h-> b bbar)_MSSM',  'Gamma(h -> b bbar)_SM'
       write(nout,313) prodxbr%gamma%tobbar(1) & 
                                   , prodxbr%gamma%tobbarSM(1)
       write(nout,'(a)')
       
       write(nout,115)  'Gamma(h-> gam gam)_MSSM',  'Gamma(h -> gam gam)_SM'
       write(nout,313) prodxbr%gamma%togamgam(1) & 
                                   , prodxbr%gamma%togamgamSM(1)      
       write(nout,'(a)')

       write(nout,115)  'Gamma(h-> tau tau)_MSSM',  'Gamma(h -> tau tau)_SM'
       write(nout,313) prodxbr%gamma%totautau(1) & 
                                   , prodxbr%gamma%totautauSM(1)      
       write(nout,'(a)')       

       write(nout,115)  'Gamma(h-> W W*)_MSSM',  'Gamma(h -> W W*)_SM'
       write(nout,313) prodxbr%gamma%toWW(1) & 
                                   , prodxbr%gamma%toWWSM(1)      
       write(nout,'(a)')  


    endif

    !    
    if (Flag%ID_predict) then

      write(nout,'(a)')'Indirect detection:'
      write(nout,'(a)')'-------------------'


      if (Flag%ID%gadiff) then
       write(nout,'(2a)') ' Diff gamma ray fluxes'//trim(dphigam_units)
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%fluxgadiff(1:n_hm)
       write(nout,'(a)')
      endif


      if (Flag%ID%gac) then
       write(nout,'(2a)') ' Gamma ray fluxes'//trim(phigam_units)
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%fluxgac(1:n_hm)
       write(nout,'(a)')
      endif

      if (Flag%ID%gam) then
       write(nout,'(2a)') ' Monocromatic gamma ray fluxes gg'//trim(phigam_units)
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%fluxgaga(1:n_hm)
       write(nout,'(a)')

       write(nout,'(2a)') ' Monocromatic gamma ray fluxes gZ'//trim(phigam_units)
       write (nout, fmt_idhm ) DSO%ID_in%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%fluxgaz(1:n_hm)
       write(nout,'(a)')
      endif


      if (Flag%ID%pb) then
       write(nout,'(a)') ' Antiproton fluxes (GeV^-1 cm^-2 s^-1 sr^-1) '
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%phipb(1:n_hm)
       write(nout,'(a)')
      endif

      if (Flag%ID%db) then
       write(nout,'(a)') ' Antideutron fluxes (GeV^-1 cm^-2 s^-1 sr^-1) '
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%phidb(1:n_hm)
       write(nout,'(a)')
      endif

      if (Flag%ID%ep) then
       write(nout,'(a)') ' Positron fluxes (GeV^-1 cm^-2 s^-1 sr^-1) ' 
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%phiep(1:n_hm)
       write(nout,'(a)')
      endif

      if (Flag%ID%epfrac) then
       write(nout,'(a)') ' Positron fluxes fraction' 
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%phiep_frac(1:n_hm)
       write(nout,'(a)')
      endif

      if (Flag%ID%musu) then
       write(nout,'(2a)') ' Muon flux from the Sun'//trim(ratenu_units)
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%ratesu(1:n_hm)
       write(nout,'(a)')
      endif

      if (Flag%ID%muea) then
       write(nout,'(2a)') ' Muon flux from the Earth'//trim(ratenu_units)
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%rateea(1:n_hm) 
       write(nout,'(a)')
      endif

      if (Flag%ID%musudiff) then
       write(nout,'(2a)') ' Diff. muon flux from the Sun'//trim(dratenu_units)
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm)
       write (nout, fmt_idhmd) DSO%ID%ratesudiff(1:n_hm)
       write(nout,'(a)')
      endif

      if (Flag%ID%mueadiff) then
       write(nout,'(2a)') ' Diff. muon flux from the Earth'//trim(dratenu_units)
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm)
       write (nout, fmt_idhmd) DSO%ID%rateeadiff(1:n_hm)
       write(nout,'(a)')
      endif


      if (Flag%ID%muha) then
       write(nout,'(a)') ' Muon flux from the Halo (km^-2 yr^-1) '
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%phimuhalo(1:n_hm)
       write(nout,'(a)')
      endif

      if (Flag%ID%muhadiff) then
       write(nout,'(a)') ' Muon flux from the Halo (km^-2 yr^-1 GeV^-1) '
       write (nout, fmt_idhm ) GIDin%hmodel(1:n_hm) 
       write (nout, fmt_idhmd) DSO%ID%phimuhalodiff(1:n_hm)
       write(nout,'(a)')
      endif


      if (Flag%ID%sigmav) then
       write(nout,'(a)') ' Netralino annihilation cross section sigma v at p=0 (cm^3 s^-1) '
       write (nout, fmt_idhmnd) DSO%ID%sigmav
       write(nout,'(a)')
      endif

      
    endif

    !
    if(DSO%Err%muSqWrongSign.eq.1 .or. DSO%Err%b.eq.1 .or. &
     DSO%Err%noRhoConvergence.eq.1 .or. DSO%Err%tachyon.eq.1 .or. & 
     DSO%Err%nonperturbative.eq.1 .or. DSO%Err%noMuConvergence.eq.1 .or. &
     DSO%Err%higgsUfb.eq.1 .or. DSO%Err%irqfp.eq.1 .or. &
     DSO%Err%badConvergence.eq.1) then
       write(nout,'(a)')'Error flags from Softsusy:'
       write(nout,'(a)')'--------------------------'
       
       write(nout,116) 'muSqWrongSign','b','noRhoConvergence','tachyon' &
            ,'nonperturbative' 
       write(nout,314) DSO%Err%muSqWrongSign, DSO%Err%b, DSO%Err%noRhoConvergence &
            ,DSO%Err%tachyon,DSO%Err%nonperturbative
       write(nout,'(a)')
       
       write(nout,117) 'noMuConvergence','higgsUfb','irqfp','badConvergence'
       write(nout,315) DSO%Err%noMuConvergence,DSO%Err%higgsUfb,DSO%Err%irqfp &
            ,DSO%Err%badConvergence
       write(nout,'(a)')
    endif
    
    if( DSO%Err%unphys.ne.0.or.DSO%Err%oh2.ne.0) then
       
       write(nout,'(a)')'Error flags from DS:'
       write(nout,'(a)')'--------------------'
       
       write(nout,118) 'LSP not neutralino','oh2_err'
       write(nout,316)  DSO%Err%unphys,DSO%Err%oh2
       write(nout,'(a)')
    endif

 100  format(3(g11.4),1x,g11.4,7x,i3)
 101  format(5(g11.4),2x,g11.4)
 102  format(3(g11.4),4x,g11.5)
 103  format(g11.4,1x,g11.4,g11.4)
 104  format(6(g11.4))
 106  format(1(g11.4))
 110  format(1x,g12.7,3x,g12.7)
 111  format(4(g12.5))
 112  format(2x,a3,4x,a16)
 113  format(2x,a14,7x,a11,6x,a11)
 114  format(2x,a10,7x,a11,6x,a11,5x,a5,8x,a12)
 115  format(2x,a30,4x,a30)
 116  format(2x,a13,3x,a5,6x,a18,6x,a5,a18)
 117  format(2x,a13,5x,a5,4x,a10,6x,a15)
 118  format(2x,a18,7x,a10)
 119  format(2x,a30,4x,a30)
 120  format(5x,a55)

 310  format(g11.5,2x,g11.5)
 311  format(2x,g11.5,9x,g11.5,7x,g11.5)
 312  format(1x,g11.5,8x,i3,13x,g11.5,5x,g11.5,4x,g11.5)
 313  format(18x,g11.5,25x,g11.5)
 314  format(5x,i3,12x,i3,12x,i3,15x,i3,12x,i3)
 315  format(5x,i3,12x,i3,11x,i3,13x,i3)
 316  format(g8.3,23x,i3)
 317  format(35x,g11.5)

 578  format(2x,a3,8x,a5,6x,a3,8x,a9,7x,a8,3x)
 581  format(2x,a5,6x,a5,6x,a7,6x,a7,4x)
 583  format(2x,a5,7x,a15,6x)
 585  format(2x,a5,6x,a5,6x,a5,6x,a7,4x)
 587  format(2x,a6,5x,a6,5x,a6,5x,a6,5x,a6,5x,a6,5x)
 590  format(2x,a6,5x,a7,5x,a5,5x,a25,6x)
 591  format(2x,a6,5x,a6,5x,a5,5x,a6,5x,a8,5x,a8,5x)
 592  format(2x,a6,5x,a6,5x,a7,4x,a7,3x,a9,4x,a9,5x)
 593  format(2x,a6,5x,a6,5x,a6,5x)
 594  format(2x,a7,4x,a7,4x,a6,5x)

 691  format(2x,a10,5x,a10,5x,a10,5x,a10,5x,a10)
 692  format(2(g15.4),3x,3(g15.4))


 694  format('//trim(numcat(' ',n_hm))//'G10.4) 

!fmt_idgadiff = '('//trim(numcat(' ',num_hm))//'E15.5)'


end subroutine WriteDebug


subroutine SetProposeMatrix
  real proj_len(num_params_used)
  real U(num_params_used,num_params_used)
  real vecs(num_slow,num_params_used)
  integer i, ii,j

  !std devs of base parameters

   if (.not. allocated(sigmas)) allocate(sigmas(num_params_used))
   do i = 1, num_params_used
     sigmas(i) = sqrt(propose_matrix(i,i))
   end do

   do i = 1, num_params_used
     propose_matrix(i,:) = propose_matrix(i,:) / sigmas(i)
     propose_matrix(:,i) = propose_matrix(:,i) / sigmas(i)
   end do

   if (num_nuis /= 0) then
      !Get the conditional covariance by projecting the inverse covariances of nuis parameters
            if (.not. allocated(propose_matrix_nuis)) then  
              allocate(propose_matrix_nuis(num_nuis, num_nuis))
              allocate(propose_diag_nuis(num_nuis))
            end if
            U = propose_matrix
            write(*,*) 'Inverting U matrix'
            call Matrix_Inverse(U)
            propose_matrix_nuis = U(num_slow+1:num_params_used, num_slow+1:num_params_used)
            call Diagonalize(propose_matrix_nuis,propose_diag_nuis, num_nuis)
            !propose_matrix^-1 = U D U^T, returning U in propose_matrix
            propose_matrix_nuis = transpose(propose_matrix_nuis)
            if (any(propose_diag_nuis <= 0)) &
              call DoStop('Nuis proposal matrix has negative or zero eigenvalues')

            propose_diag_nuis = 1/sqrt(propose_diag_nuis)
   end if

   if (num_slow /= 0) then
  
   if (.not. allocated(propose_diag)) allocate(propose_diag(num_params_used))
   call Diagonalize(propose_matrix, propose_diag, num_params_used)
       !propose_matrix = U D U^T, returning U in propose_matrix

   if (any(propose_diag <= 0)) &
        call DoStop('Proposal matrix has negative or zero eigenvalues')
   propose_diag = sqrt(max(1e-12,propose_diag))

!Get projected lengths 
   do i = 1, num_params_used
          vecs(:,i) = propose_diag(i)*propose_matrix(1:num_slow,i)
          proj_len(i) = sum(vecs(:,i)**2)  
   end do

       if (.not. allocated(slow_evecs)) allocate(slow_evecs(num_slow))
!keep evectors with longest projected lengths in slow dimensions, orthogonal to previous longest       
       do i = 1, num_slow
         j = MaxIndex(proj_len, num_params_used)
         slow_evecs(i) = j
         do ii= 1, num_params_used
           if (proj_len(ii) /= 0. .and. ii/=j) then
            vecs(:,ii) = vecs(:,ii) - sum(vecs(:,j)*vecs(:,ii))*vecs(:,j)/proj_len(j)
                  !Take out projection onto jth eigendirection
            proj_len(ii) = sum(vecs(:,ii)**2)
           end if
         end do
         proj_len(j) = 0.
       end do

   end if

end subroutine SetProposeMatrix

subroutine SetProposeMatrixOLD
  real U(num_params_used,num_params_used)
  real vecs(num_hard,num_params_used)
  integer i, ii,j

!std devs of base parameters

   if (.not. allocated(sigmas)) allocate(sigmas(num_params_used))
   do i = 1, num_params_used
     sigmas(i) = sqrt(propose_matrix(i,i))
   end do

   do i = 1, num_params_used
     propose_matrix(i,:) = propose_matrix(i,:) / sigmas(i)
     propose_matrix(:,i) = propose_matrix(:,i) / sigmas(i)
   end do
   
  
   if (.not. allocated(propose_diag)) allocate(propose_diag(num_params_used))
   call Diagonalize(propose_matrix, propose_diag, num_params_used)
   !propose_matrix = U D U^T, returning U in propose_matrix
   if (any(propose_diag <= 0)) &
        stop 'Proposal matrix has negative or zero eigenvalues'
   propose_diag = sqrt(max(1e-12,propose_diag))
         
end subroutine SetProposeMatrixOLD

subroutine DoStop(S)
 character(LEN=*), intent(in), optional :: S
 integer ierror
        if (present(S)) write (*,*) trim(S)
#ifdef MPI
        MPI_StartTime = MPI_WTime() - MPI_StartTime 
        if (Feedback > 0 .and. MPIRank==0) then
           write (*,*) 'Total time:', nint(MPI_StartTime), &
                                   '(',MPI_StartTime/(60*60),' hours)'
           write (*,*) 'Fct evaluations: ', num
        end if
        call mpi_finalize(ierror)
#endif
        stop

end subroutine DoStop
   
end module ParamDef
